/* Canvas.c
 *
 * Copyright 1999, Paul Kienzle (pkienzle@cs.indiana.edu)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * * Initial revision
 */

#define VERBOSE 0
#include "stickers.h"
#include "canvas.h"
#include "misc.h"

CanvasCursor cursor = {};
Canvas canvas = {};

/* ============= Picture cursor ======================================== */
/* I am using the current sticker as the cursor for the sticker drawing
 * area.  Unfortunately, X doesn't support colour cursors, so we have to
 * simulate the cursor by drawing the sticker on the page where the cursor
 * should be, then removing it and drawing it again whenever we move the
 * cursor.  To implement remove, we use a backing store to keep track of
 * what was under the sticker before we move it.  Note that we must draw
 * the new cursor on the page prior to updating the screen to avoid flicker
 * since the new cursor may overlap the old.
 *
 * To use, do the following:
 *      cursor_clear()
 *      do changes to canvas/cursor
 *      cursor_draw()
 *      update screen with changes outside old/new cursor
 *
 * Note: I tried implementing this with a child window, but gdk doesn't
 * support the CWBackingStore attribute on child window creation, so
 * I got lots of flicker.  Using a toplevel window worked a bit better,
 * but clipping at the canvas boundaries seemed like it would be painful
 * to implement.
 */
void cursor_clear()
{
  GdkGC *gc = canvas.widget->style->fg_gc[GTK_WIDGET_STATE(canvas.widget)];
  TRACE("Clearing cursor.");

  /* Reset image on canvas with cursor backing store. */
  if (cursor.old_drawn) {
    gdk_draw_pixmap(canvas.canvas, gc, cursor.backing_store,
		    0, 0, cursor.old_left, cursor.old_top, 
		    cursor.old_width, cursor.old_height);
  }
}

void cursor_draw(gboolean on_canvas, gint x, gint y)
{
  gboolean draw_new;
  gint top, left;
  GdkGC *gc = canvas.widget->style->fg_gc[GTK_WIDGET_STATE(canvas.widget)];

  TRACE("Drawing cursor.");

  cursor.x = x; cursor.y = y; cursor.on_canvas = on_canvas;
  draw_new = on_canvas && cursor.image;
  TRACE("draw new=%d, cursor.image=%p", draw_new, cursor.image);

  /* Place hotspot in the middle of the sticker */
  left = x - cursor.hot_x;
  top = y - cursor.hot_y;
  
  /* Draw the new cursor */
  if (draw_new) {
    /* Make sure the backing store exists and is of the correct size. */
    if (cursor.backing_store == NULL) {
      cursor.backing_store = gdk_pixmap_new(canvas.widget->window, 
					    cursor.width, cursor.height, -1);
    }
    
    /* Set cursor backing store with image on canvas. */
    gdk_draw_pixmap(cursor.backing_store, gc, canvas.canvas,
		    left, top, 0, 0, cursor.width, cursor.height);

    /* Place the cursor on the canvas. */
    gdk_imlib_paste_image(cursor.image, canvas.canvas, 
			  left, top, cursor.width, cursor.height);
  }

  /* Update the screen where the old cursor was */
  if (cursor.old_drawn)
    gdk_draw_pixmap(canvas.widget->window, gc, canvas.canvas,
		    cursor.old_left, cursor.old_top,
		    cursor.old_left, cursor.old_top,
		    cursor.old_width, cursor.old_height);

  /* Update the screen where the new cursor is, and remember its location */
  if (draw_new) {
    gdk_draw_pixmap(canvas.widget->window, gc, canvas.canvas, 
		    left, top, left, top, cursor.width, cursor.height);
    cursor.old_left = left;
    cursor.old_top = top;
    cursor.old_width = cursor.width;
    cursor.old_height = cursor.height;
  }
  cursor.old_drawn = draw_new;
  TRACE("draw done");
}


void cursor_set(GdkImlibImage *image, gint hot_x, gint hot_y)
{
  TRACE("");
  cursor_clear();
  cursor.hot_x = hot_x;
  cursor.hot_y = hot_y;
  if (!cursor.image || 
      cursor.width != image->rgb_width || 
      cursor.height != image->rgb_height) {
    if (cursor.backing_store) gdk_pixmap_unref(cursor.backing_store);
    cursor.backing_store = NULL;
    cursor.width = image->rgb_width;
    cursor.height = image->rgb_height;
  }
  cursor.image = image;
  cursor_draw(cursor.on_canvas, cursor.x, cursor.y);
}

void cursor_off()
{
  TRACE("");
  cursor_clear();
  cursor_draw(FALSE, cursor.x, cursor.y);
}

void cursor_on()
{
  TRACE("");
  cursor_clear();
  cursor_draw(TRUE, cursor.x, cursor.y);
}
  

/* =========== Canvas Window ============================================= */


/* When creating the canvas widget, create a null cursor to display on it */
static void canvas_map(GtkWidget *widget, GdkEvent *event)
{
  TRACE("Mapping canvas");

  /* Set the invisible cursor */
  if (config.use_tiny_cursor)
    gdk_window_set_cursor(widget->window, config.tiny_cursor);
  confine_add(widget);
}

static void canvas_unmap(GtkWidget *widget, GdkEvent *event)
{
  TRACE("Unmapping canvas");
  if (config.confine) confine_remove(widget);
}

/* Resize messages for sticker drawing area. */
static gint canvas_configure(GtkWidget *widget, GdkEventConfigure *event)
{
  TRACE("Configuring canvas.");

  /* Create a new backing pixmap of the appropriate size (preserving
     aspect ratio). */
  if (canvas.background) {
    gdouble x_ratio = (gdouble)widget->allocation.width
      / (gdouble)canvas.background->rgb_width;
    gdouble y_ratio = (gdouble)widget->allocation.height
      / (gdouble)canvas.background->rgb_height;
    canvas.scale = x_ratio < y_ratio ? x_ratio : y_ratio;
  }
  else
    canvas.scale = 1.0;
TRACE("canvas scale=%f", canvas.scale);
  canvas.width = widget->allocation.width;
  canvas.height = widget->allocation.height;
  if (canvas.canvas) gdk_pixmap_unref(canvas.canvas);
  canvas.canvas = gdk_pixmap_new(widget->window, 
				 canvas.width, canvas.height, -1);

  scene_redraw();
  return TRUE;
}

static gint canvas_destroy(GtkWidget *widget, GdkEvent *event)
{
  TRACE("canvas destroy");
  if (canvas.canvas) gdk_pixmap_unref(canvas.canvas);
  canvas.canvas = NULL;
  return TRUE;
}

/* Expose messages for sticker drawing area. */
static gint canvas_expose (GtkWidget *widget, GdkEventExpose *event)
{
  TRACE("Exposing canvas.");

  /* Redraw the exposed area from the backing pixmap */
  gdk_draw_pixmap(widget->window, 
		  widget->style->fg_gc[GTK_WIDGET_STATE(widget)], 
		  canvas.canvas,
		  event->area.x, event->area.y,
		  event->area.x, event->area.y,
		  event->area.width, event->area.height);

  return FALSE;
}

/* Press message: paste sticker onto canvas where cursor is. */
static gint canvas_press (GtkWidget *widget, GdkEventButton *event)
{
  TRACE("Canvas Press.");

  /* Keep the cursor on the canvas while the button is pressed. */
  if (config.confine && config.paint) 
    confine_pointer(widget);

  scene_push();

  return TRUE;
}

/* Release message: stop pasting stickers. */
static gint canvas_release (GtkWidget *widget, GdkEventButton *event)
{
  TRACE("Canvas release.");

  /* Allow the cursor to leave the canvas when the button is released,
     but still keep it in the application. */
  if (config.confine && config.paint) 
    confine_pointer(gtk_widget_get_toplevel(widget));
  return TRUE;
}


/* Move message: Change the cursor position. */
static gint canvas_move (GtkWidget *widget, GdkEventMotion *event)
{
  gint x, y;
  GdkModifierType state;

  TRACE("Canvas move");

  /* This is directly from scribble-simple --- I don't understand is_hint. */
  if (event->is_hint) 
    gdk_window_get_pointer(widget->window, &x, &y, &state);
  else {
    x = event->x;
    y = event->y;
    state =  event->state;
  }


  cursor_clear();
  cursor.x = x;
  cursor.y = y;

  /* If a button is pressed, place the sticker on the canvas. */
  if (state&(GDK_BUTTON1_MASK|GDK_BUTTON2_MASK|GDK_BUTTON3_MASK)) 
    if (config.paint) scene_push();
    else cursor_draw(cursor.on_canvas, x, y);
  else cursor_draw(cursor.on_canvas, x, y);

  return TRUE;
}

/* Enter notify: draw the cursor on the canvas. */
static gint canvas_enter (GtkWidget *widget, GdkEventCrossing *event)
{
  GdkModifierType state;
  gint x,y;

  TRACE("Entering canvas.");
  cursor_clear();
  gdk_window_get_pointer (widget->window, &x, &y, &state);
  cursor_draw(TRUE, x, y);

  return TRUE;
}

/* Leave notify: remove the cursor from the canvas. */
static gint canvas_leave (GtkWidget *widget, GdkEventCrossing *event)
{
  TRACE("Leaving canvas.");

  /* Clear the cursor and redraw the canvas without one. */
  cursor_clear();
  cursor_draw(FALSE, cursor.x, cursor.y);

  /* Release the cursor to root (required to change the mouse pointer back
     to the application default). */
  if (config.confine) confine_pointer(gtk_widget_get_toplevel(widget));

  return TRUE;
}

/* ================================================================== */
/* Exported canvas functions. */

/* Define the canvas: called from main at the start of the program. */
GtkWidget *canvas_new(GdkImlibImage *image)
{
  GtkWidget *frame;
  int w, h;

  TRACE("Canvas new.");

  canvas.canvas = NULL;
  canvas.widget = gtk_drawing_area_new ();
  canvas.background = image;
  cursor.on_canvas = FALSE;
  cursor.image = NULL;
  cursor.backing_store = NULL;
  canvas.cursor_stack = NULL;
  canvas.cursor_stack_size = 0;
  if (image) w=image->rgb_width,h=image->rgb_height;
  else w=540, h=420;

  if (canvas.background) {
    frame = gtk_aspect_frame_new(NULL, 0.5, 0.5, (gdouble)w/(gdouble)h, FALSE);
    gtk_container_add(GTK_CONTAINER(frame), canvas.widget);
  }
  else
    frame = canvas.widget;

  gtk_drawing_area_size (GTK_DRAWING_AREA (canvas.widget), w, h);

  gtk_signal_connect (GTK_OBJECT (canvas.widget), "destroy_event",
		      (GtkSignalFunc) canvas_destroy, NULL);
  gtk_signal_connect (GTK_OBJECT (canvas.widget), "expose_event",
		      (GtkSignalFunc) canvas_expose, NULL);
  gtk_signal_connect (GTK_OBJECT (canvas.widget),"configure_event",
		      (GtkSignalFunc) canvas_configure, NULL);
  gtk_signal_connect (GTK_OBJECT (canvas.widget), "button_press_event",
		      (GtkSignalFunc) canvas_press, NULL);
  gtk_signal_connect (GTK_OBJECT (canvas.widget), "button_release_event",
		      (GtkSignalFunc) canvas_release, NULL);
  gtk_signal_connect (GTK_OBJECT (canvas.widget), "motion_notify_event",
		      (GtkSignalFunc) canvas_move, NULL);
  gtk_signal_connect (GTK_OBJECT (canvas.widget), "enter_notify_event",
		      (GtkSignalFunc) canvas_enter, NULL);
  gtk_signal_connect (GTK_OBJECT (canvas.widget), "leave_notify_event",
		      (GtkSignalFunc) canvas_leave, NULL);
  gtk_signal_connect (GTK_OBJECT (canvas.widget), "map_event",
		      (GtkSignalFunc) canvas_map, NULL);
  gtk_signal_connect (GTK_OBJECT (canvas.widget), "unmap_event",
		      (GtkSignalFunc) canvas_unmap, NULL);

  gtk_widget_set_events (canvas.widget, GDK_EXPOSURE_MASK
			 | GDK_LEAVE_NOTIFY_MASK
			 | GDK_ENTER_NOTIFY_MASK
			 | GDK_BUTTON_PRESS_MASK
			 | GDK_BUTTON_RELEASE_MASK
			 | GDK_POINTER_MOTION_MASK
			 | GDK_POINTER_MOTION_HINT_MASK);

  return frame;
}


/* Set the initial image. */
void canvas_clear(GtkWidget *widget)
{
  TRACE("Setting initial image as %p", canvas.background);

  cursor.old_drawn = FALSE;
  if (canvas.background)
    gdk_imlib_paste_image(canvas.background, canvas.canvas, 0, 0, 
			  canvas.width, canvas.height);
  else
    gdk_draw_rectangle (canvas.canvas,
			widget->style->white_gc,
			TRUE, 0, 0, canvas.width, canvas.height);
}

GdkImlibImage* canvas_image()
{
  GdkImlibImage *image;
  TRACE("Printing canvas.");

  /* Note: available in Imlib 1.9.4 */
  image = gdk_imlib_create_image_from_drawable(canvas.canvas, NULL, 0, 0, 
					 canvas.width, canvas.height);
  return image;
}
