/* Main.c
 *
 * Copyright 1999, Paul Kienzle (pkienzle@cs.indiana.edu)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * * Nov 15, 1999 - PAK && Cees de Groot
 *   include math.h
 *   cycle sticker set back to start when reaching the end, both for
 *   mouse and keyboard operations
 * * July 25, 1999 - PAK
 *   allow for an empty scene directory
 * * May 29, 1999 - PAK
 *   remove borders from windows (for less flash when using sticker controls)
 *   preserve sticker state when cycling through a sticker set
 *   allow single step simultaneous keys (e.g., for diagonal motion)
 * * May 13, 1999 - PAK
 *   add break to the GDK_Undo switch so that undo key works
 *   kill timer before destroying the window (otherwise it segv's)
 * * Initial revision
 */

// #include <gdk/gdkx.h>
#include <gdk/gdkkeysyms.h>
// #include <gdk/gdkcursors.h>
#include <math.h>
#include <stdlib.h>

#define VERBOSE 0  /* Whether to print trace messages */
#define USE_TABS 1 /* Whether sticker sets show folder tabs or not */
#include "stickers.h"
#include "misc.h"
#include "canvas.h"

typedef struct {
  GList *history, *history_tail;
  gint history_size;
  GList *stickers;
  GList *extras;
  GList *sticker_radio;
  GList *sticker_notebooks;
  GtkNotebook *current_notebook;
  GdkImlibImage *background;
  gchar *file;
} Scene;

typedef void (*ControlType)(gint);
typedef enum {
  CONTROL_OFF = 0,
  BIDIRECTIONAL,
  FORWARD,
  REVERSE
} ControlDirection;
typedef struct CONTROL
{
  GHashTable* table;
  gboolean reverse;
  gboolean restore;
  gboolean fine;
  gboolean timer_running;
  gboolean initial_keys_signalled;
  gboolean initial_keys_gathered;
  guint32 timer;
} Control;

static Scene scene;
static Control control = {};
Config config;

/* XXX FIXME XXX
there's got to be a better way to avoid unnecessary sticker rendering
and screen redaws than having this global flag running around. */
static int _processing = 0;

static guint control_hash(gpointer d)
{
  return (gint)d;
}
static void count_controls (ControlType f, ControlDirection d, gint *count)
{
  (*count)++;
}
static void do_controls (ControlType f, ControlDirection *d)
{
  switch(*d) {
  case BIDIRECTIONAL:
    (*f)((control.restore?0:(control.fine?1:2))
	 * (control.reverse?-1:1) );
    break;
  case FORWARD:
    (*f)(control.restore?0:(control.fine?1:2));
    break;
  case REVERSE:
    (*f)(control.restore?0:(control.fine?-1:-2));
    break;
  case CONTROL_OFF:
    break;
  }
}
static gint control_timer(gpointer data)
{
  g_return_val_if_fail(control.timer!=0, FALSE);

  if (!control.timer_running) {
    TRACE("timer switching to spin delay");
    control.timer = gtk_timeout_add(config.delay, control_timer, NULL);
    control.timer_running = TRUE;
    g_hash_table_foreach(control.table, (GHFunc)do_controls, NULL);
    sticker_set_cursor();
    return FALSE;
  }
  else {
    g_hash_table_foreach(control.table, (GHFunc)do_controls, NULL);
    sticker_set_cursor();
    return TRUE;
  }
}
static gint gather_initial_keys(gpointer data)
{
  control.initial_keys_gathered = TRUE;
  g_hash_table_foreach(control.table, (GHFunc)do_controls, NULL);
  sticker_set_cursor();
  return FALSE;
}

void control_start(ControlType f, ControlDirection d)
{
  ControlDirection *data;
  TRACE("control start");

  data = g_hash_table_lookup(control.table, f);
  if (data) g_free(data);
  data = g_new(ControlDirection, 1);
  *data = d;
  g_hash_table_insert(control.table, f, data);
  if (!control.timer) {
    TRACE("starting timer");
    if (!control.initial_keys_signalled && config.initial_delay > 150)
      gtk_timeout_add(100, gather_initial_keys, NULL);
    control.initial_keys_signalled = TRUE;
    control.timer = gtk_timeout_add(config.initial_delay, control_timer, NULL);
  }
  if (control.initial_keys_gathered && !control.timer_running) {
    do_controls(f, data);
    sticker_set_cursor();
  }
}
void control_stop(ControlType f, ControlDirection d)
{
  ControlDirection *data;
  int total = 0;

  TRACE("control stop");
  /* Note: pressing the key for FORWARD before releasing the REVERSE
   * should cause the motion to go forward (it does since g_hash_table_insert
   * overrides the old value in the hash table). Releasing the REVERSE
   * should not stop the FORWARD motion since the FORWARD key is still
   * pressed. So before removing the control from the table, make sure
   * that the control we are removing is in the correct direction. */
  data = g_hash_table_lookup (control.table, f);
  if (data && *data == d) {
    /* check for fast press/release */
    if (!control.initial_keys_gathered) do_controls(f, data); 
    g_free(data);
    g_hash_table_remove(control.table, f);
    g_hash_table_foreach(control.table, (GHFunc)count_controls, &total);
    if (total == 0) {
      TRACE("stopping timer");
      if (control.timer) gtk_timeout_remove(control.timer);
      control.timer = 0;
      control.timer_running = FALSE;
      control.initial_keys_gathered = FALSE;
      control.initial_keys_signalled = FALSE;
    }
  }
}


void gtk_notebook_next_page_cycle (GtkNotebook *notebook)
{
  int current;
  
  current = gtk_notebook_get_current_page(notebook);
  gtk_notebook_next_page(notebook);
  if (gtk_notebook_get_current_page(notebook) == current)
    gtk_notebook_set_page(notebook, 0);
}

void gtk_notebook_prev_page_cycle (GtkNotebook *notebook)
{
  int current;
  
  current = gtk_notebook_get_current_page(notebook);
  gtk_notebook_prev_page(notebook);
  if (gtk_notebook_get_current_page(notebook) == current)
    gtk_notebook_set_page(notebook, g_list_length(notebook->children)-1);
}

int size2pixel(int percent)
{
  return (percent * gdk_screen_width() + 50) / 100;
}

/* Create a picture widget of the given dimension while preserving the
 * aspect ratio. */
GtkWidget *picture_new(GdkImlibImage *image, int size)
{
  GdkPixmap *pixmap;
  GdkBitmap *mask;
  gint w, h;

  /* convert size in % of screen to size in pixels */
  size = size2pixel(size);
  if (size == 0) size = 1;
  if (image->rgb_width > image->rgb_height)
    w = size, h = (size*image->rgb_height)/image->rgb_width;
  else
    w = (size*image->rgb_width)/image->rgb_height, h=size;
  gdk_imlib_render(image, w, h);
  pixmap = gdk_imlib_move_image(image);
  mask = gdk_imlib_move_mask(image);
  return gtk_pixmap_new(pixmap, mask);
}

GtkWidget* button_new_with_image(gchar *imagename, gint size, gchar *label)
{
  GdkImlibImage *image;
  GtkWidget *button;
  
  if (fexist(imagename)) image = gdk_imlib_load_image(imagename);
  else image = NULL;
  if (image != NULL) {
    button = gtk_button_new();
    gtk_container_add (GTK_CONTAINER(button), picture_new(image, size));
    gdk_imlib_kill_image(image);
  }
  else {
    button = gtk_button_new_with_label(label);
  }
  return button;
}


/* Button interface grabs the pointer whenever it is clicked, which
 * means that if you leave the button before releasing the pointer,
 * your next click has no effect. Should really "unpress" the button
 * instead (don't "release" it though since that will call the button
 * action to take effect!). Note that the button_down flag seems to
 * indicate whether the button should be drawn as pressed on reenter. 
 */
static void leave_button (GtkWidget *widget)
{
  GTK_BUTTON(widget)->button_down = 0;
  gtk_grab_remove(widget);
}


/* == Application Window ========================================== */

/* Set the this-may-take-a-while mouse pointer */
void set_pointer_wait (GtkWidget *widget)
{
  GtkWidget *toplevel = gtk_widget_get_toplevel(widget);
  GdkCursor *pointer = gdk_cursor_new(GDK_WATCH);

  TRACE("widget=%p, root=%p", widget, toplevel);

  if (widget != toplevel)
    gdk_window_set_cursor (widget->window, pointer);
  gdk_window_set_cursor (toplevel->window, pointer);
  gdk_cursor_destroy(pointer);
}

/* Set the usual mouse pointer */  
void set_pointer (GtkWidget *widget)
{
  GtkWidget *toplevel = gtk_widget_get_toplevel(widget);
  GdkCursor *pointer = gdk_cursor_new (GDK_LEFT_PTR);

  TRACE("widget=%p, root=%p", widget, toplevel);

  if (widget != toplevel) {
    if (config.use_tiny_cursor)
      gdk_window_set_cursor (canvas.widget->window, config.tiny_cursor);
    else if (config.use_big_cursor)
      gdk_window_set_cursor (canvas.widget->window, config.big_cursor);
    else
      gdk_window_set_cursor (canvas.widget->window, pointer);
  }
  if (config.use_big_cursor)
    gdk_window_set_cursor (toplevel->window, config.big_cursor);
  else
    gdk_window_set_cursor (toplevel->window, pointer);
  gdk_cursor_destroy(pointer);
}

void map_window (GtkWidget *widget, GdkEvent *event)
{

  TRACE("Mapping window.");

  /* Make sure the cursor can't escape from the application */
  if (config.confine) confine_pointer(widget);
}

void enter_window (GtkWidget *widget, GdkEventCrossing *event)
{

  /* Make sure the cursor can't escape from the application */
  if (event->detail != GDK_NOTIFY_INFERIOR) {
    TRACE("Entering window");
    if (config.confine) confine_pointer(widget);
    gdk_key_repeat_disable();
  }
}

void leave_window (GtkWidget *widget, GdkEventCrossing *event)
{
  if (event->detail != GDK_NOTIFY_INFERIOR) {
    TRACE("Leaving window");
    /* Cursor escaped despite our best efforts, release it's confinement. */
    if (config.confine) gdk_pointer_ungrab(GDK_CURRENT_TIME);
    gdk_key_repeat_restore();
  }
}


/* == Sticker buttons =============================================== */
/* Switch stickers by clicking on the button. I'm simulating the
 * effect of a radio button here by shutting off the old button and
 * turning on the new. This works fine except when the old button is
 * equal to the new. In this case, clicking it turns it off, so I
 * manually have to turn it on again to make sure that at least one
 * sticker is selected.   */

static void sticker_button_clicked (GtkToggleButton *this_button, Sticker *s)
{
  GtkToggleButton *button;
  GtkNotebook *notebook;
  GList *list;

  /* Very obscure: as the sticker notebooks are destroyed, they seem to
   * switch pages (presumably removing the page which is about to be
   * destroyed causes it to switch to another page). A page switch
   * causes the sticker it contains to be activated. Activating the
   * sticker requires accessing it. Now, the sticker notebooks are
   * only destroyed after the scene is destroyed, so the stickers are
   * deleted by the time we get here. So, check if the stickers still
   * exist before processing the sticker button click. */
  if (scene.extras==NULL) return;

  TRACE("sticker button clicked %s", s->image->filename);

  if (this_button->active) {
    /* Make all other buttons inactive */
    for (list = scene.sticker_radio; list; list = list->next) {
      button = list->data;
      if (button->active && button != this_button)
	gtk_toggle_button_set_state(button, FALSE);
    }

    if (sticker) {
      /* Save the x and y location of the last sticker */
      s->x = sticker->x;
      s->y = sticker->y;
      if (scene.current_notebook
	  == GTK_NOTEBOOK(GTK_WIDGET(this_button)->parent)) {
	/* If last sticker was in the same sticker set as the current,
	   preserve all the sticker attributes */
	  s->hue = sticker->hue;
	  s->saturation = sticker->saturation;
	  s->brightness = sticker->brightness;
	  s->alpha = sticker->alpha;
	  s->scale = sticker->scale;
	  s->rotate = sticker->rotate;
	  s->flip = sticker->flip;
      }
    }
    /* Make the sticker active */
    if (_processing) sticker_set(s);
    scene.current_notebook = GTK_NOTEBOOK(GTK_WIDGET(this_button)->parent);
  }
  else {
    /* Make sure some other button is active */
    for (list = scene.sticker_radio; list; list = list->next) {
      button = list->data;
      if (button->active && button != this_button) break;
    }

    /* If nothing else active, go to the next notebook page and activate it. */
    if (list == NULL) {
      notebook = GTK_NOTEBOOK(GTK_WIDGET(this_button)->parent);
      gtk_notebook_next_page_cycle(notebook);
    }
  }

}
static void sticker_button_destroy(GtkWidget *widget)
{
  scene.sticker_radio = g_list_remove(scene.sticker_radio, widget);
}

static void switch_page(GtkWidget *widget, GtkNotebookPage *page, int pagenum)
{
  GtkToggleButton *button;


  TRACE("switch to page %d", pagenum);
  button = GTK_TOGGLE_BUTTON(page->child);
  gtk_toggle_button_set_state(button, TRUE);
}

void notebook_destroy(GtkWidget*widget)
{
  TRACE("notebook destroy");
  scene.sticker_notebooks = g_list_remove(scene.sticker_notebooks, widget);
}

static GtkWidget * add_set(GList *list)
{
#if USE_TABS
  static guchar tab_rgb[3]={0};
  static GdkImlibImage *tab = NULL;
#endif
  GtkNotebook *notebook;
  GtkWidget *button;

  TRACE("add set");
  notebook = GTK_NOTEBOOK(gtk_notebook_new ());
  scene.sticker_notebooks = g_list_append(scene.sticker_notebooks, notebook);
  gtk_signal_connect (GTK_OBJECT (notebook), "destroy",
		      GTK_SIGNAL_FUNC (notebook_destroy), NULL); 
  gtk_signal_connect (GTK_OBJECT (notebook), "switch_page",
		      GTK_SIGNAL_FUNC (switch_page), NULL);

  gtk_notebook_set_show_border(notebook, FALSE);
#if USE_TABS
  if (tab == NULL) {
    GdkImlibColor transparency;
    tab = gdk_imlib_create_image_from_data(tab_rgb,NULL,1,1);
    transparency.r=transparency.g=transparency.b=0;
    gdk_imlib_set_image_shape(tab, &transparency);
  }
  gtk_notebook_set_scrollable(notebook, TRUE);
  gtk_notebook_set_show_tabs(notebook, TRUE);
  gtk_notebook_set_tab_pos(notebook, GTK_POS_BOTTOM);
  gtk_notebook_popup_enable(notebook);
  gtk_notebook_set_show_border(notebook, 0);
  gtk_notebook_set_tab_border(notebook, 0);
#else
  gtk_notebook_set_scrollable(notebook, FALSE);
  gtk_notebook_set_show_tabs(notebook, FALSE);
#endif
  while (list) {
    Sticker *s = list->data;

    button = gtk_toggle_button_new();
    scene.sticker_radio = g_list_prepend(scene.sticker_radio, button);
    gtk_signal_connect (GTK_OBJECT (button), "clicked",
			GTK_SIGNAL_FUNC (sticker_button_clicked),s);
    gtk_signal_connect (GTK_OBJECT (button), "leave",
			GTK_SIGNAL_FUNC (leave_button), NULL);
    gtk_signal_connect (GTK_OBJECT (button), "destroy",
			GTK_SIGNAL_FUNC (sticker_button_destroy), NULL);
    gtk_container_add (GTK_CONTAINER(button), 
		       picture_new(s->image, config.button_size));
#if USE_TABS
    gtk_notebook_append_page_menu(notebook, button,
				  picture_new(tab, 0),
				  picture_new(s->image, config.button_size));
#else
    gtk_notebook_append_page(notebook, button, NULL);
#endif
    list = g_list_next(list);
  }

  /*  gtk_notebook_set_page(notebook, 0); */
  return GTK_WIDGET(notebook);
}

GtkWidget * add_stickers()
{
  GtkWidget *table, *event;
  gint length, across, rows, i, j;
  GList *list;

  TRACE("add stickers");
  /* Balancing magic: find the number of buttons which will fit across
  * the screen and the number of buttons which you have to display.
  * Divide these (truncating upward) to get the number of rows required. 
  * Again take the number of buttons which you have to display, but now 
  * divide it by the number of rows required (truncating downward) to 
  * get the minimum number of buttons per row. The remaining length%rows
  * buttons should be distributed on a first come first serve basis. */
  across = 100 / config.button_size;
  length = g_list_length(scene.stickers);
  rows = (length+across-1)/across;
  across = length/rows;
  rows = length%rows;
  table = gtk_table_new(rows, across+(rows?1:0), TRUE);


  /* Now do the list */  
  list = scene.stickers;
  for (j=0; list; j++) {
    for (i = 0; i < across + (rows>0?1:0); i++) {
      gtk_table_attach (GTK_TABLE (table), add_set(list->data),
			i, i+1, j, j+1, GTK_FILL, GTK_FILL, 0, 0);
      list = g_list_next(list);
    }
    rows--;
  }

  /* Create an event box which we can hang tool tips on. */
  event = gtk_event_box_new();
  gtk_container_add( GTK_CONTAINER(event), table );
  config_tip(event, _("\
Click image to select sticker. Click again to select the \
next sticker in the set. Use \"[\" or \"]\" to cycle through the stickers in \
the current set. Right-click on tabs to see all stickers in the set at once. \
Use \"P\" or \"p\" to cycle through the available sets.\nClick in \
scene to paste the sticker, or use the arrow keys to \
to position the sticker and Insert to paste it on the scene. Use \
Ctrl for finer cursor control."));

  return event;
}

void scene_next_sticker_in_set(gint control)
{
  TRACE("Next sticker");
  if (scene.current_notebook == NULL) return;
  if (control > 0)
    gtk_notebook_next_page_cycle(scene.current_notebook);
  else
    gtk_notebook_prev_page_cycle(scene.current_notebook);
}

void scene_next_sticker_set(gint control)
{
  GtkNotebook *notebook;
  GtkToggleButton *button;
  GList *next;
  
  TRACE("Next set");
  if (control > 0) {
    next = g_list_next(g_list_find(scene.sticker_notebooks,
				   scene.current_notebook));
    if (next == NULL) next = scene.sticker_notebooks;
  }
  else if (control < 0) {
    next = g_list_previous(g_list_find(scene.sticker_notebooks,
				       scene.current_notebook));
    if (next == NULL) next = g_list_last(scene.sticker_notebooks);
  }
  else return;
  notebook = next->data;
  if (notebook->cur_page) {
    button = GTK_TOGGLE_BUTTON(notebook->cur_page->child);
    gtk_toggle_button_set_state(button, TRUE);
  }
  else
    gtk_notebook_set_page(notebook, 0);
}





/* === Sticker Control Area ============================================ */
/* Sticker controls adjust the current sticker. Since it is nice to see
what you are adjusting, they show the current sticker as they adjust
it. To avoid flash when going from control to control, we group all
the controls in one box (called the sticker_controlarea), and display
the current sticker whenever we enter the box. The sticker could be
displayed in a popup window or some such, but it was more convenient
just to draw it in the middle of the canvas, especially since it isn't
clear how to size/resize the window when the sticker size or rotation
are changed. Too much is ugly, and not enough doesn't show any
sticker. */

static void sticker_controlarea_enter(GtkWidget *widget, 
				      GdkEventCrossing *event)
{
  if (event->detail != GDK_NOTIFY_INFERIOR) {
    cursor_clear();
    cursor_draw(TRUE, canvas.width/2, canvas.height/2);
  }
}
static void sticker_controlarea_leave(GtkWidget *widget, 
				      GdkEventCrossing *event)
{
  if (event->detail != GDK_NOTIFY_INFERIOR) {
    cursor_clear();
    cursor_draw(FALSE, cursor.x, cursor.y);
  }
}

GtkWidget* sticker_controlarea_new(GtkWidget **box)
{
  GtkWidget* event_box = gtk_event_box_new();
  gtk_signal_connect (GTK_OBJECT (event_box), "enter_notify_event",
		      (GtkSignalFunc) sticker_controlarea_enter, NULL);
  gtk_signal_connect (GTK_OBJECT (event_box), "leave_notify_event",
		      (GtkSignalFunc) sticker_controlarea_leave, NULL);
  *box = gtk_vbox_new (FALSE, 0);
  gtk_container_add(GTK_CONTAINER(event_box), *box);
  return event_box;
}

/* == Scene controls ================================================== */

typedef struct {
  ControlType control_type;
  ControlDirection direction;
  gboolean started;
} StickerControl;
static gint sticker_control_press (GtkWidget *widget, 
				   GdkEventButton *event,
				   StickerControl *data)
{
  TRACE("sticker control press");

  /* Doesn't matter which button is pressed. Could instead make
   * different buttons do different things, like button 1 for fast,
   * button 3 for slow or reverse and button 2 for reset */
  control_start(data->control_type, data->direction);
  data->started = TRUE;
  return TRUE;
}
static gint sticker_control_release (GtkWidget *widget,
				     GdkEventButton *event,
				     StickerControl *data)
{
  control_stop(data->control_type, data->direction);
  data->started = FALSE;
  return TRUE;
}
static void sticker_control_leave (GtkWidget *widget, StickerControl *data)
{
  TRACE("sticker control leave");
  /* Fix me: this works for the other buttons, why not these? */
  gtk_grab_remove(widget);
  if (data->started) control_stop(data->control_type, data->direction);
}
static void sticker_control_destroy (GtkWidget *widget, StickerControl *data)
{
  TRACE("sticker control destroy");
  g_free(data);
}

GtkWidget *add_sticker_control(GtkWidget *control_area, char *icon, 
			       ControlType f, ControlDirection d, gchar *tip)
{
  StickerControl *data;
  gchar *imagename;
  GtkWidget *button;

  TRACE("add sticker control %s", icon);

  imagename = g_strconcat(config.icon_dir, "icon_", icon, ".xpm", NULL);
  button = button_new_with_image(imagename, config.button_size, icon); 
  gtk_box_pack_start (GTK_BOX (control_area), button, 
		      FALSE, FALSE, 0);

  data = g_new(StickerControl, 1);
  data->control_type = f;
  data->direction = d;
  data->started = FALSE;
  gtk_signal_connect_after (GTK_OBJECT (button), "button_press_event",
		      GTK_SIGNAL_FUNC (sticker_control_press), data);
  gtk_signal_connect (GTK_OBJECT (button), "button_release_event",
		      GTK_SIGNAL_FUNC (sticker_control_release), data);
  gtk_signal_connect (GTK_OBJECT (button), "leave",
		      GTK_SIGNAL_FUNC (sticker_control_leave), data);
  gtk_signal_connect (GTK_OBJECT (button), "destroy",
		      GTK_SIGNAL_FUNC (sticker_control_destroy), data);

  if (tip) config_tip(button, tip);
  return button;
}

GtkWidget *
add_control(GtkWidget *control_box, char *icon, void (*callback)(), gchar *tip)
{
  gchar *imagename;
  GtkWidget *button;

  TRACE("add control %s", icon);
  imagename = g_strconcat(config.icon_dir, "icon_", icon, ".xpm", NULL);
  button = button_new_with_image(imagename, config.button_size, icon);
  g_free(imagename);
  gtk_box_pack_start (GTK_BOX (control_box), button, FALSE, FALSE, 0);
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
		      GTK_SIGNAL_FUNC (callback), NULL);
  gtk_signal_connect (GTK_OBJECT (button), "leave",
		      GTK_SIGNAL_FUNC (leave_button), NULL);
  if (tip) config_tip(button, tip);
  return button;
}

typedef struct TOGGLE {
  gboolean *state;
  GdkImlibImage *on_image, *off_image;
  GtkWidget *surface;
} Toggle;
void toggle_clicked(GtkWidget *button, Toggle *toggle)
{
  GdkImlibImage *image;
  GdkPixmap *pixmap;
  GdkBitmap *mask;
  gint w, h, size;

  TRACE("toggle clicked");
  *(toggle->state) = !*(toggle->state);
  if (toggle->surface == NULL) return;

  if (*(toggle->state)) image = toggle->on_image;
  else image = toggle->off_image;

  size = size2pixel(config.button_size);
  if (image->rgb_width > image->rgb_height) 
    w = size, h = (size*image->rgb_height)/image->rgb_width;
  else
    w = (size*image->rgb_width)/image->rgb_height, h = size;
  gdk_imlib_render(image, w, h);
  pixmap = gdk_imlib_move_image(image);
  mask = gdk_imlib_move_mask(image);

  /* XXX FIXME XXX : do I need to delete the previous surface image? */
  gtk_pixmap_set(GTK_PIXMAP(toggle->surface), pixmap, mask);
}
static void toggle_destroy(GtkButton *button, Toggle *toggle)
{
  TRACE("toggle destroy");
  if (toggle->on_image) gdk_imlib_destroy_image(toggle->on_image);
  if (toggle->off_image) gdk_imlib_destroy_image(toggle->off_image);
  g_free(toggle);
}
GtkWidget *add_toggle(GtkWidget *control_box, gchar *icon_on,
		      gchar *icon_off, gboolean *state, gchar *tip)
{
  GtkWidget *button;
  gchar *imagename;
  Toggle *toggle = g_new(Toggle, 1);

  TRACE("add toggle");
  toggle->state = state;

  imagename = g_strconcat(config.icon_dir, "icon_", icon_on, ".xpm", NULL);
  if (fexist(imagename)) toggle->on_image = gdk_imlib_load_image(imagename);
  else toggle->on_image = NULL;
  g_free(imagename);

  imagename = g_strconcat(config.icon_dir, "icon_", icon_off, ".xpm", NULL);
  if (fexist(imagename)) toggle->off_image = gdk_imlib_load_image(imagename);
  else toggle->off_image = NULL;
  g_free(imagename);

  if (toggle->on_image==NULL || toggle->off_image==NULL) {
    char *label = g_strconcat(icon_on, "/", icon_off, NULL);
    button = gtk_toggle_button_new_with_label(label);
    gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(button), *state);
    g_free(label);
    toggle->surface = NULL;
  }
  else {
    button = gtk_button_new();
    toggle->surface = picture_new(*state?toggle->on_image:toggle->off_image,
				  config.button_size);
    gtk_container_add (GTK_CONTAINER (button), toggle->surface);
  }
  gtk_box_pack_start (GTK_BOX (control_box), button, FALSE, FALSE, 0);
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
		      GTK_SIGNAL_FUNC (toggle_clicked), toggle);
  gtk_signal_connect (GTK_OBJECT (button), "leave",
		      GTK_SIGNAL_FUNC (leave_button), NULL);
  gtk_signal_connect (GTK_OBJECT (button), "destroy",
		      GTK_SIGNAL_FUNC (toggle_destroy), toggle);
  if (tip) config_tip(button, tip);
  return button;
}

/* == Scene functions ==================================================== */

static void push_one_sticker(Sticker* new, GdkImlibImage *image)
{
  GdkGC *gc = canvas.widget->style->fg_gc[GTK_WIDGET_STATE(canvas.widget)];
  Sticker *s;
  GdkRectangle rect;

  /* Account for the size of the new sticker */
  scene.history_size += new->width * new->height * 3 + sizeof(Sticker);

  /* Remove from bottom of stack until within requested limits. */
  while (scene.history && scene.history_size>config.max_history*100000) {
    GList *new_tail;

    new_tail = g_list_previous(scene.history_tail);
    scene.history = g_list_remove_link(scene.history, scene.history_tail);
    s = scene.history_tail->data;
    gdk_pixmap_unref(s->backing_store);
    scene.history_size -= s->width * s->height * 3 + sizeof(Sticker);
    g_free(s);
    g_list_free_1(scene.history_tail);
    scene.history_tail = new_tail;
  }
  if (!scene.history) scene.history_tail = NULL;

  /* Add new sticker to the head of the list */
  s = g_new(Sticker, 1);
  memcpy(s, new, sizeof(Sticker));
  scene.history = g_list_prepend(scene.history, s);
  if (!scene.history_tail) scene.history_tail = scene.history;

  /* Grab the image off the canvas. */
  rect.x = s->x - s->width/2;
  rect.y = s->y - s->height/2;
  rect.width = s->width;
  rect.height = s->height;
  s->backing_store = gdk_pixmap_new(canvas.widget->window, 
				    rect.width, rect.height, -1);
  gdk_draw_pixmap(s->backing_store, gc, canvas.canvas,
		  rect.x, rect.y, 0, 0, rect.width, rect.height);

  /* Paste the new sticker on the canvas. */
  gdk_imlib_paste_image(image, canvas.canvas, 
			rect.x, rect.y, rect.width, rect.height); 
}

void scene_redraw(void)
{
  GList *list;

  TRACE("scene redraw");
  if (!_processing) return;

  set_pointer_wait (canvas.widget);

  /* Clear the canvas and redraw the history. */
  canvas_clear();
  list = g_list_reverse(scene.history);
  scene.history = scene.history_tail = NULL;
  while (list) {
    GList *next;
    GdkImlibImage *image;
    Sticker *s = list->data;

    if (s->backing_store) gdk_pixmap_unref(s->backing_store);
    image = sticker_render(s);
    s->x = rint(s->rel_x * canvas.width);
    s->y = rint(s->rel_y * canvas.height);
    push_one_sticker(s, image);
    gdk_imlib_kill_image(image);
    next = g_list_next(list);
    g_free(s);
    g_list_free_1(list);
    list = next;
  }
  cursor_draw(cursor.on_canvas, cursor.x, cursor.y);
  gtk_widget_draw(canvas.widget, NULL);

  /* Reset the cursor in case the image scale has changed */
  sticker_set_cursor();

  set_pointer (canvas.widget);
}

void scene_push()
{
  Sticker *s;
  GdkRectangle rect;

  TRACE("scene push");
  if (!sticker) return;
  sticker->x = cursor.x;
  sticker->y = cursor.y;
  sticker->rel_x = (gdouble)cursor.x/(gdouble)canvas.width;
  sticker->rel_y = (gdouble)cursor.y/(gdouble)canvas.height;

  /* Check for duplicate sticker (same image, same place) */
  if (scene.history) {
    s = scene.history->data;
    if (sticker_match(s, sticker)) return;
  }
  cursor_clear();
  push_one_sticker(sticker, cursor.image);
  cursor_draw(cursor.on_canvas, cursor.x, cursor.y);
  rect.x = sticker->x-sticker->width/2;
  rect.y = sticker->y-sticker->height/2;
  rect.width = sticker->width;
  rect.height = sticker->height;
  gtk_widget_draw(canvas.widget, &rect);
}

void scene_pop()
{
  GdkGC *gc = canvas.widget->style->fg_gc[GTK_WIDGET_STATE(canvas.widget)];
  GdkRectangle rect;
  Sticker *top;

  TRACE("Clearing last sticker.");
  if (scene.history) {
    /* Remove first item from the history list. */
    top = scene.history->data;
    scene.history = g_list_remove(scene.history, top);
    if (scene.history == NULL) scene.history_tail = NULL;

    /* Account for the size of the removed sticker */
    scene.history_size -= top->width * top->height * 3 + sizeof(Sticker);

    /* Restore the image on the canvas */
    cursor_clear();
    rect.x = top->x - top->width/2;
    rect.y = top->y - top->height/2;
    rect.width = top->width;
    rect.height = top->height;
    gdk_draw_pixmap(canvas.canvas, gc, top->backing_store, 
		    0, 0, rect.x, rect.y, rect.width, rect.height);
    cursor_draw(cursor.on_canvas, cursor.x, cursor.y);
    gtk_widget_draw(canvas.widget, &rect);

    /* Free the sticker */
    gdk_pixmap_unref(top->backing_store);
    g_free(top);
  }
}

/* Clear all the stickers off the scene. */
void scene_clear()
{
  Sticker *top;

  TRACE("Clearing canvas!");
  while (scene.history) {
    /* Remove first item from the history list. */
    top = scene.history->data;
    scene.history = g_list_remove(scene.history, top);

    /* Free the sticker */
    gdk_pixmap_unref(top->backing_store);
    g_free(top);
  }
  scene.history_size = 0;
  scene.history_tail = NULL;
  canvas_clear();
  gtk_widget_draw(canvas.widget, NULL);
}

/* Find the sticker under the cursor, pull it out of the list and make
 * it the current sticker */
void scene_grab()
{
}

static void scene_save_image(gboolean print)
{
  GdkImlibImage *image, *thumb;
  gchar *path, *ext;
  gchar base[256];
  gint w, h, l, t;
  gchar *sig = config_signature();

  /* Write the signature on the canvas (if any) and grab the canvas image */
  if (sig[0]) {
    GdkFont *font;
    GdkGC *gc = canvas.widget->style->fg_gc[GTK_WIDGET_STATE(canvas.widget)];

    /* Draw signature text on canvas */
    font = NULL;
    if (config.sigfont[0]) font = gdk_font_load(config.sigfont);
    if (font == NULL) font = gdk_font_ref(canvas.widget->style->font);
    w = gdk_string_width (font, sig)+1;
    h = font->ascent + font->descent+1;
    l = canvas.width-w;
    t = canvas.height-h;
    gdk_draw_text(canvas.canvas, font, gc, 
		  l, t+font->ascent, sig, strlen(sig));
    gdk_font_unref(font);

    /* Grab canvas image */
    image = canvas_image();

    /* Restore canvas */
    gdk_draw_pixmap(canvas.canvas, gc, canvas.widget->window, 
		    l, t, l, t, w, h);
  }
  else image = canvas_image();
  g_free(sig);

  /* Save the image as a thumbnail */
  if (canvas.width > canvas.height)
    w = 100, h = (100*canvas.height)/canvas.width;
  else
    w = (100*canvas.width)/canvas.height, h = 100;
  thumb = gdk_imlib_clone_scaled_image(image, w, h);
  path = g_strconcat(config.work_dir, strrchr(scene.file,'/')+1, ".xpm", NULL);
  gdk_imlib_save_image(thumb, path, NULL);
  gdk_imlib_kill_image(thumb);
  g_free(path);

  /* Save the image as config.printfile (if any) and print (if needed) */
  if (config.printfile[0]) {
    ext = strrchr(scene.file,'.');
    *ext = '\0'; /* <blah>.scene -> <blah> */
    snprintf(base, sizeof(base), config.printfile, strrchr(scene.file,'/')+1);
    *ext = '.'; /* <blah> -> <blah>.scene */
    path = g_strconcat(config.work_dir, base, NULL);
    gdk_imlib_save_image(image, path, NULL);
    gdk_imlib_kill_image(image);

    if (print && config.print && config.printcmd[0]) {
      snprintf(base, sizeof(base), config.printcmd, path);
      system(base);
    }
    g_free(path);
  }
}

void scene_print(void)
{
  scene_save_image(TRUE);
}

/* Save the scene to a file */
static void save_available(Sticker *sticker, FILE *file)
{
  fprintf(file, " %s", 
	  strrchr(sticker->image->filename,'/')+1);
  if (sticker->base_scale != 1.0)
    fprintf(file, " : %f", sticker->base_scale); 
}
void scene_save()
{
  GList *next;
  gint size;
  FILE *scenefile;
  gchar *path;

  TRACE("scene save");
  /* Don't save if you are modifying the scenes in the scene directory */
  if (strcmp(config.work_dir, config.sticker_dir) == 0) return;

  /* Don't save empty scenes */
  if (!scene.history) return;

  /* Write the scene image */
  scene_save_image(FALSE);

  /* Open the scene file */
  path = g_strconcat(config.work_dir, strrchr(scene.file,'/')+1, NULL);
  scenefile = fopen(path, "w");
  if (scenefile == NULL) {
    printf(_("Could not open scene file %s for writing\n"), path);
    g_free(path);
    return;
  }
  g_free(path);

  /* Write the scene file */
  if (scene.background) {
    fprintf(scenefile, "@background %s  d %d,%d\n", 
	    strrchr(scene.background->filename,'/')+1, 
	    canvas.width, canvas.height);
  }
  for (next=scene.stickers; next; next = g_list_next(next)) {
    fprintf(scenefile, "@stickers ");
    g_list_foreach(next->data, (GFunc)save_available, scenefile);
    fprintf(scenefile, "\n");
  }
  size = 0;
  for (next=scene.history; g_list_next(next); next=g_list_next(next)) {
    Sticker *s = next->data;
    size += s->width*s->height*3 + sizeof(Sticker);
    if (size > config.max_saved*100000) break;
  }
  while (next) {
    fprintf(scenefile, "@put ");
    sticker_save(next->data, scenefile);
    next = g_list_previous(next);
  }
  fclose(scenefile);
}

static Sticker *new_or_existing_sticker(gchar *name)
{
  GdkImlibImage *image;
  gchar *path;
  GList *list;

  TRACE("new or existing sticker %s",name);

  /* Find name of file containing the image */
  path = g_strconcat(config.sticker_dir, name, NULL);

  /* Locate sticker in complete list */
  for (list = scene.extras; list; list = g_list_next(list)) {
    if (strcmp(path, ((Sticker *)(list->data))->image->filename) == 0) {
      g_free(path);
      return list->data;
    }
  }

  /* Doesn't exist already, so load its image */
  if (fexist(path)) image = gdk_imlib_load_image(path);
  else image=NULL;

  /* If loaded, add to complete list */
  if (image == NULL) {
    printf(_("Could not load sticker image %s\n"), path);
    g_free(path);
    return NULL;
  }
  else {
    Sticker *s = sticker_new(image);
    scene.extras = g_list_prepend(scene.extras, s);
    g_free(path);
    return s;
  }
}

/* Load the scene from a file */
gboolean scene_load(char *filename)
{
  gchar token[256];
  gint w, h, v;
  FILE *file;
  gchar *imagename;

  TRACE("scene load");

  file = fopen(filename, "r");
  if (file == NULL) {
    printf(_("Could not open scene file %s.\n"), filename);
    return FALSE;
  }

  /* Initialize scene structure */
  memset(&scene, 0, sizeof(Scene));

  v = fscanf(file, " @ %255s ", token);
  if (v != 1) {
    printf(_("%s doesn't appear to be a scene file %s\n"), filename);
    fclose(file);
    return FALSE;
  }
  scene.file = g_strdup(filename);
  
  while (v == 1) {
    if (strcmp(token, "background") == 0) {
      v = fscanf(file, " %255s d %d , %d  ", token, &w, &h);
      if (v < 1) printf(_("Background image not specified.\n"));
      if (scene.background == NULL) {
	imagename = g_strconcat(config.sticker_dir, token, NULL);
	scene.background = gdk_imlib_load_image(imagename);
	if (scene.background == NULL)
	  printf(_("Could not load background image file %s\n"), 
		 imagename);
	g_free(imagename);
      }
      v = fscanf(file, " @ %255s ", token);
    }
    else if (strcmp(token, "stickers") == 0) {
      GList *list = NULL;
      for (;;) {
	Sticker *s;
	gdouble scale;

	v = fscanf(file, " @ %255s ", token);
	if (v == 1 || v==-1) break;
	v = fscanf(file, " %255s : %lf ", token, &scale);
	if (v > 0) {
	  s = new_or_existing_sticker(token);
	  if (s) {
	    if (v > 1) s->base_scale = scale;
	    TRACE("%d: base scale for %s is %f\n", v, token, s->base_scale);
	    list = g_list_prepend(list, s);
	  }
	}
      }
      if (list) scene.stickers =  g_list_prepend(scene.stickers, 
						 g_list_reverse(list));
    }
    else if (strcmp(token, "put") == 0) {
      for (;;) {
	Sticker *put;
	gchar *imagename;

	v = fscanf(file, " @ %255s ", token);
	if (v == 1 || v==-1) break;
	put = sticker_load(file, &imagename);
	if (put) {
	  Sticker *s;
	  s = new_or_existing_sticker(imagename);
	  if (s) {
	    put->image = s->image;
	    put->base_scale = s->base_scale;
	    scene.history = g_list_prepend(scene.history, put);
	  }
	  else g_free(put);
	}
	else printf(_("Unable to load sticker %s\n"), imagename);
      }
    }
    else {
      printf(_("Unidentified token %s in file %s\n"), token, filename);
      fclose(file);
      return TRUE;
    }
  }
  if (v != -1) {
    fscanf(file, " %255s ", token);
    printf(_("Unrecognized string in file %s: %s\n"),filename,token);
  }
  fclose(file);
  scene.stickers = g_list_reverse(scene.stickers);
  return scene.stickers != NULL;
}


/* Quit the scene. */
void scene_destroy (GtkWidget *widget)
{
  _processing = 0;

  TRACE("scene destroy");
  if (control.timer) gtk_timeout_remove(control.timer);
  
  g_hash_table_destroy(control.table);
  while (scene.stickers) {
    GList *l;
    for (l = scene.stickers->data; l; l = g_list_remove(l, l->data)) ;
    scene.stickers = g_list_remove(scene.stickers, scene.stickers->data);
  }
  while (scene.extras) {
    Sticker *top = scene.extras->data;
    gdk_imlib_destroy_image(top->image);
    scene.extras = g_list_remove(scene.extras, top);
    g_free(top);
  }
  while (scene.history) {
    Sticker *top = scene.history->data;
    scene.history = g_list_remove(scene.history, top);
    gdk_pixmap_unref(top->backing_store);
    g_free(top);
  }
  if (scene.background) gdk_imlib_destroy_image(scene.background);
  select_scene();
}

void scene_config (GtkWidget *widget)
{
  GtkWidget *toplevel = gtk_widget_get_toplevel(widget);

  /* Hide the window, edit the configuration, then show the window again. */
  gtk_widget_hide(toplevel);
  config_edit(); 
  gtk_widget_show(toplevel);

  /* Reset the pointer in case pointer type has changed */
  set_pointer(canvas.widget);
}

void scene_exit (GtkWidget *widget)
{
  TRACE("scene_exit");
  scene_save();

  gtk_widget_destroy(gtk_widget_get_toplevel(widget));
}
/* Process keystrokes. */
static int get_keystroke (GtkWidget * widget, GdkEventKey * event)
{
  if (event->type == GDK_KEY_PRESS) 
    TRACE("Pressing key %d", event->keyval);
  else TRACE("Releasing key %d", event->keyval);

  if (event->type == GDK_KEY_PRESS) switch(event->keyval) {
    /* Modifiers */
  case GDK_Alt_L: case GDK_Alt_R:
    control.restore = TRUE; break;
  case GDK_Shift_L: case GDK_Shift_R:
    control.reverse = TRUE; break;
  case GDK_Control_L: case GDK_Control_R:
    control.fine = TRUE; break;

    /* Things to do immediately */
  case GDK_Escape:
    scene_exit(widget); break;
  case GDK_m: case GDK_M: 
    sticker_flip(); break;
  case GDK_Home:
    sticker_restore(); break;
  case GDK_End:
    scene_clear(); break;
  case GDK_exclam:
    scene_config(widget); break;
  case GDK_numbersign:
    if (config.print) scene_print();
    break;

    /* Things to do continuously */
  case GDK_Insert:
    control_start((ControlType)scene_push, BIDIRECTIONAL); break;
  case GDK_Undo: case GDK_BackSpace:case GDK_Delete:
    control_start((ControlType)scene_pop, FORWARD); break;
  case GDK_p: case GDK_P: 
    control_start(scene_next_sticker_set, BIDIRECTIONAL); break;
  case GDK_bracketleft: case GDK_braceleft:
    control_start(scene_next_sticker_in_set, REVERSE); break;
  case GDK_bracketright: case GDK_braceright:
    control_start(scene_next_sticker_in_set, FORWARD); break;
  case GDK_Up:            
    control_start(sticker_vertical, REVERSE); break;
  case GDK_Down:          
    control_start(sticker_vertical, FORWARD); break;
  case GDK_Right:         
    control_start(sticker_horizontal, FORWARD); break;
  case GDK_Left:          
    control_start(sticker_horizontal, REVERSE); break;
  case GDK_r: case GDK_R: 
    control_start(sticker_rotate, BIDIRECTIONAL); break;
  case GDK_c: case GDK_C: 
  case GDK_h: case GDK_H: 
    control_start(sticker_hue, BIDIRECTIONAL); break;
  case GDK_s: case GDK_S: 
    control_start(sticker_saturation, BIDIRECTIONAL); break;
  case GDK_b: case GDK_B: 
    control_start(sticker_brightness, BIDIRECTIONAL); break;
  case GDK_f: case GDK_F: 
    control_start(sticker_fade, BIDIRECTIONAL); break;
  case GDK_Next:          
    control_start(sticker_scale, REVERSE); break;
  case GDK_Prior:         
    control_start(sticker_scale, FORWARD); break;
  }

  else switch(event->keyval) { /* key release */
    /* Modifiers */
  case GDK_Alt_L: case GDK_Alt_R:
    control.restore = FALSE; break;
  case GDK_Shift_L: case GDK_Shift_R:
    control.reverse = FALSE; break;
  case GDK_Control_L: case GDK_Control_R:
    control.fine = FALSE; break;


    /* Things to do continuously */
  case GDK_Insert:
    control_stop((ControlType)scene_push, BIDIRECTIONAL); break;
  case GDK_Undo: case GDK_BackSpace: case GDK_Delete:
    control_stop((ControlType)scene_pop, FORWARD); break;
  case GDK_p: case GDK_P: 
    control_stop(scene_next_sticker_set, BIDIRECTIONAL); break;
  case GDK_bracketleft: case GDK_braceleft:
    control_stop(scene_next_sticker_in_set, REVERSE); break;
  case GDK_bracketright: case GDK_braceright:
    control_stop(scene_next_sticker_in_set, FORWARD); break;
  case GDK_Up:            
    control_stop(sticker_vertical, REVERSE); break;
  case GDK_Down:          
    control_stop(sticker_vertical, FORWARD); break;
  case GDK_Right:         
    control_stop(sticker_horizontal, FORWARD); break;
  case GDK_Left:
    control_stop(sticker_horizontal, REVERSE); break;
  case GDK_r: case GDK_R: 
    control_stop(sticker_rotate, BIDIRECTIONAL); break;
  case GDK_c: case GDK_C: 
  case GDK_h: case GDK_H: 
    control_stop(sticker_hue, BIDIRECTIONAL); break;
  case GDK_s: case GDK_S: 
    control_stop(sticker_saturation, BIDIRECTIONAL); break;
  case GDK_b: case GDK_B: 
    control_stop(sticker_brightness, BIDIRECTIONAL); break;
  case GDK_f: case GDK_F: 
    control_stop(sticker_fade, BIDIRECTIONAL); break;
  case GDK_Next:          
    control_stop(sticker_scale, REVERSE); break;
  case GDK_Prior:         
    control_stop(sticker_scale, FORWARD); break;
  }

  if (event->type == GDK_KEY_PRESS)
    gtk_signal_emit_stop_by_name(GTK_OBJECT(widget), "key_press_event");
  else
    gtk_signal_emit_stop_by_name(GTK_OBJECT(widget), "key_release_event");

  return FALSE;
}

void scene_new()
{
  GtkWidget *window;
  GtkWidget *drawing_area;
  GtkWidget *table;
  GtkWidget *box;

  TRACE("scene new");
  /* Create the toplevel window */
  control.table = g_hash_table_new((GHashFunc)control_hash, NULL);

  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title (GTK_WINDOW (window), _("Sticker Book"));
  gtk_window_set_policy (GTK_WINDOW (window), TRUE, TRUE, TRUE);
  if (config.maximize)
    gtk_widget_set_usize (window, gdk_screen_width(), gdk_screen_height());

  gtk_window_position (GTK_WINDOW (window), GTK_WIN_POS_CENTER);
  gtk_signal_connect (GTK_OBJECT (window), "destroy",
		      GTK_SIGNAL_FUNC (scene_destroy), NULL);
  gtk_signal_connect (GTK_OBJECT (window), "key_press_event",
		      GTK_SIGNAL_FUNC (get_keystroke), NULL);
  gtk_signal_connect (GTK_OBJECT (window), "key_release_event",
		      GTK_SIGNAL_FUNC (get_keystroke), NULL);
  gtk_signal_connect (GTK_OBJECT (window), "enter_notify_event",
		      (GtkSignalFunc) enter_window, NULL);
  gtk_signal_connect (GTK_OBJECT (window), "leave_notify_event",
		      (GtkSignalFunc) leave_window, NULL);
  gtk_signal_connect (GTK_OBJECT(window),"map_event",
		      (GtkSignalFunc) map_window, NULL);
  gtk_signal_connect (GTK_OBJECT(window),"unmap_event",
		      (GtkSignalFunc) map_window, NULL);
  gtk_widget_set_events (window, GDK_EXPOSURE_MASK
			 | GDK_LEAVE_NOTIFY_MASK
			 | GDK_ENTER_NOTIFY_MASK
			 | GDK_KEY_PRESS_MASK
			 | GDK_KEY_RELEASE_MASK);

  /* sets the border width of the window. */
  gtk_container_border_width (GTK_CONTAINER (window), 0);

  /* Put a table into the window */
  table = gtk_table_new (3, 2, FALSE);
  gtk_container_border_width (GTK_CONTAINER (table), 0);
  gtk_container_add (GTK_CONTAINER (window), table);

  /* Add the drawing area */
  drawing_area = canvas_new(scene.background);
  gtk_table_attach (GTK_TABLE (table), drawing_area,
		    1, 2, 0, 1, 
		    GTK_FILL|GTK_SHRINK|GTK_EXPAND,
		    GTK_FILL|GTK_SHRINK|GTK_EXPAND,
		    0, 0);

  /* Add the stickers */
  gtk_table_attach (GTK_TABLE (table), add_stickers(),
		    0, 3, 1, 2,
		    0, 0, 0, 0);
  sticker = NULL;

  /* Add the scene controls */
  box = gtk_vbox_new (FALSE, 0);
  gtk_table_attach (GTK_TABLE (table), box,
		    2, 3, 0, 1,
		    0, GTK_FILL|GTK_SHRINK|GTK_EXPAND,
		    0, 0);

  add_control(box, "Undo", scene_pop, _("\
Remove the last sticker from the screen. Equivalent to Backspace or Delete."));

#if 0
  add_control(box, "Grab", scene_grab, _("\
Grab a sticker from the scene, unstick it, and make it the current sticker."));
#endif

  add_control(box, "Clear", scene_clear, _("\
Clear the scene. Equivalent to End."));

  add_toggle(box, "Paint", "Stamp", &config.paint, _("\
Toggle paint/stamp mode. In paint mode, click and drag with the mouse \
to place a sequence of stickers. In stamp mode, need to click each place \
you want a sticker. No keyboard equivalent."));

  add_control(box, "Stop", scene_exit, _("\
Save scene and exit. Equivalent to Esc."));

  if (config.config) add_control(box, "Configure", scene_config, _("\
Go to configure window. Equivalent to \"!\"."));

  if (config.print) add_control(box, "Print", scene_print, _("\
Print scene. Equivalent to \"#\". You must configure the working \
directory, the saved image file and the print command before printing. \
Use \"!\" to enter the configuration window where you can set these values."));


  /* Add the sticker controls */
  gtk_table_attach (GTK_TABLE (table), sticker_controlarea_new(&box),
		    0, 1, 0, 1,
		    0, GTK_FILL|GTK_SHRINK|GTK_EXPAND,
		    0, 0);

  add_sticker_control(box, "Grow", sticker_scale, FORWARD, _("\
Enlarge sticker. Equivalent to PgUp.\nUse Ctrl \
for finer control. Use Alt to restore the normal size."));

  add_sticker_control(box, "Shrink", sticker_scale, REVERSE, _("\
Shrink sticker. Equivalent to PgDn.\nUse Ctrl \
for finer control. Use Alt to restore the normal size."));

  add_sticker_control(box, "Rotate", sticker_rotate, BIDIRECTIONAL, _("\
Rotate sticker clock-wise. Equivalent to \"r\".\nUse Ctrl for finer control. \
Use Shift to shift direction. Use Alt to restore to upright."));

  add_control(box, "Mirror", sticker_flip, _("\
Mirror image of sticker. Equivalent to \"m\"."));

  add_sticker_control(box, "Hue", sticker_hue, BIDIRECTIONAL, _("\
Change sticker hue. This cycles through the colours of the rainbow from \
red to violet then the cycle repeats. Equivalent to \"h\" or \"c\" for \
colour.\nUse Ctrl for finer control. Use Shift to shift direction. Use \
Alt to restore the original sticker hues."));

  add_sticker_control(box, "Saturation", sticker_saturation, BIDIRECTIONAL,
_("\
Change sticker saturation. This makes the colour richer until it reaches a \
maximum, then makes it paler again until the sticker is in black and white, \
then the cycle repeats. Equivalent to \"s\".\nUse Ctrl for finer control. \
Use Shift to shift direction. Use Alt to restore the original saturation."));

  add_sticker_control(box, "Brightness", sticker_brightness, BIDIRECTIONAL, 
		      _("\
Change sticker brightness. This removes the shading on the sticker until \
it is completely bright (but not white---you must desaturate it to get \
white), then shades it again until it is completely black, then the cycle \
repeats. Equivalent to \"b\".\nUse Ctrl for finer control. Use shift to \
shift direction. Use Alt to restore the original brightness."));

#if 0
  add_sticker_control(box, "Fade", sticker_fade, BIDIRECTIONAL, _("\
Fade sticker in and out. This makes the sticker more and more see-through \
until it is invisible, then makes it solid again. Equivalent to \"f\".\nUse \
Ctrl for finer steps. Use Shift to reverse direction. Use Alt to restore \
the original transparency."));
#endif

  add_control(box, "Restore", sticker_restore, _("\
Restore sticker size, colour and orientation. Equivalent to Home."));

#if 0
  gchar *audiofile = gstrconcat(sticker_dir, string, ".wav", NULL);
  playfile(audiofile);
  g_free(audiofile);
#endif

  gtk_widget_show_all (window);

  /* Now that we have everything configured, we can turn on sticker
     processing, draw the scene, set the sticker and get on with it. */
  _processing = !0; 
  scene_redraw(); /* Note: sets the mouse pointer, so we don't have to */
  scene_next_sticker_set(1);
}


/* == Main ======================================================== */
int main (int argc, char *argv[])
{
  setlocale (LC_ALL, "");
  bindtextdomain (PACKAGE, LOCALEDIR);
  textdomain (PACKAGE);
  gtk_init (&argc, &argv);
  gdk_imlib_init();

  config_load(argc==2?argv[1]:NULL);
  select_scene();
  gtk_main ();

  return 0;
}
