/* Misc.c
 *
 * Copyright 1999, Paul Kienzle (pkienzle@cs.indiana.edu)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * * Initial revision
 */

#define VERBOSE 0
#include <stdlib.h>
#include <math.h>
#include "stickers.h"
#include "misc.h"

/* ===================================================================== */
/* Generic routines which ought to be in a library somewhere. */
#include <sys/stat.h>
int fexist(char *name)
{
  struct stat buf;

  return lstat(name, &buf) == 0;
}

/* A routine that ought to be in the gdk library: build a cursor from
 * a text layout with '1' for foreground, '0' for background and ' '
 * for transparent.
 */
GdkCursor *gdk_cursor_new_from_data(const gchar *bits[], 
				    gint width, gint height,
				    GdkColor *fg, GdkColor *bg, 
				    gint hot_x, gint hot_y)
{
  GdkBitmap *bitmap, *mask;
  GdkCursor *cursor;
  guchar *data, *ptr;
  gint i, j;

  /* Though it does not say so on the X11 manual pages, the bitmap
   * format consists of scan lines padded to byte boundaries with the
   * bits in "reverse" order within each byte (lo bits come before hi
   * bits in the bitmap).  I assume this representation is platform
   * independent.  Let me know if it doesn't work for you. */

  /* Create space for the bitmap, padding the scanlines to byte boundaries. */
  data = g_new(guchar, ((width+7)/8)*height);

  /* Build bitmap */
  ptr = data;
  for (i=0; i < height; i++) {
    for (j=0; j < width; j++) {
      *ptr = (*ptr >> 1)|(bits[i][j]=='1'?0x80:0);
      if (j%8 == 7) ptr++;
    }
    if (j%8) *ptr++ >>= 8-j%8;
  }
  bitmap = gdk_bitmap_create_from_data(NULL, data, width, height);

  /* Build mask */
  ptr = data;
  for (i=0; i < height; i++) {
    for (j=0; j < width; j++) {
      *ptr = (*ptr >> 1)|(bits[i][j]==' '?0:0x80);
      if (j%8 == 7) ptr++;
    }
    if (j%8) *ptr++ >>= 8-j%8;
  }
  mask = gdk_bitmap_create_from_data(NULL, data, width, height);

  /* Build cursor from bitmap and mask */
  cursor = gdk_cursor_new_from_pixmap(bitmap, mask, /* Image and mask */
				      fg, bg, /* colors */ 
				      hot_x, hot_y); /* Hot point */

  /* No longer need bitmap or mask */
  gdk_pixmap_unref(bitmap);
  gdk_pixmap_unref(mask);
  g_free(data);

  return cursor;
}


#if 0
WARNING! THIS CODE HAS NEVER BEEN COMPILED OR TESTED;

/* Routine to find a free colour in an image, presumably so you can assign
 * it to transparency. It takes black if it is available, or if every
 * possible colour is used. */
#include <string.h>
#include <stdlib.h>
static int compare_color(unsigned char *rgb1, unsigned char *rgb2)
{
  if (*rgb1 < *rgb2) return -1;
  if (*rgb1++ > *rgb2++) return 1;
  if (*rgb1 < *rgb2) return -1;
  if (*rgb1++ > *rgb2++) return 1;
  if (*rgb1 < *rgb2) return -1;
  if (*rgb1++ > *rgb2++) return 1;
  return 0;
}

void gdk_imlib_find_unused_color(GdkImlibImage *im, GdkImlibColor *c)
{
  gint size = im->rgb_widht*im->rgb_height;
  guchar *colors = g_malloc(3*im->rgb_width*im->rgb_height);
  gint i;
  guchar *cp;

  memcpy(colors, im->rgb_data, size*3);
  qsort(colors, size, 3, compare_color);
  cp = colors+3;
  old=colors[0]<<16 + colors[1]<<8 + colors[2];
  if (old > 0) {
    c->r = c->g = c->b = 0;
    g_free(colors);
    return;
  }
  for(i=1; i < size; i++) {
    new=cp[0]<<16 + cp[1]<<8 + cp[2];
    if (old>=new+1) old=new;
    else {
      c->r = (old+1)>>16;
      c->g = ((old+1)>>8)&0xFF;
      c->b = (old+1)&0xFF;
      break;
    }
    cp += 3;
  }
  if (i == size) c->r = c->g = c->b = 0;

  g_free(colors);
}
#endif


/* Code for rgb_to_hsv and hsv_to_rgb taken directly from gtkcolorsel.c
 *
 * GTK - The GIMP Toolkit
 * Copyright (C) 1995-1997 Peter Mattis, Spencer Kimball and Josh MacDonald
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */ 
static void 
hsv_to_rgb (gdouble  h, gdouble  s, gdouble  v,
	    gdouble *r, gdouble *g, gdouble *b)
{
  gint i;
  gdouble f, w, q, t;

  if (s == 0.0)
    s = 0.000001;

  if (h == -1.0)
    {
      *r = v;
      *g = v;
      *b = v;
    }
  else
    {
      if (h == 360.0)
	h = 0.0;
      h = h / 60.0;
      i = (gint) h;
      f = h - i;
      w = v * (1.0 - s);
      q = v * (1.0 - (s * f));
      t = v * (1.0 - (s * (1.0 - f)));

      switch (i)
	{
	case 0:
	  *r = v;
	  *g = t;
	  *b = w;
	  break;
	case 1:
	  *r = q;
	  *g = v;
	  *b = w;
	  break;
	case 2:
	  *r = w;
	  *g = v;
	  *b = t;
	  break;
	case 3:
	  *r = w;
	  *g = q;
	  *b = v;
	  break;
	case 4:
	  *r = t;
	  *g = w;
	  *b = v;
	  break;
	case 5:
	  *r = v;
	  *g = w;
	  *b = q;
	  break;
	}
    }
}

static void
rgb_to_hsv (gdouble  r, gdouble  g, gdouble  b,
	    gdouble *h, gdouble *s, gdouble *v)
{
  double max, min, delta;

  max = r;
  if (g > max)
    max = g;
  if (b > max)
    max = b;

  min = r;
  if (g < min)
    min = g;
  if (b < min)
    min = b;

  *v = max;

  if (max != 0.0)
    *s = (max - min) / max;
  else
    *s = 0.0;

  if (*s == 0.0)
    *h = -1.0;
  else
    {
      delta = max - min;

      if (r == max)
	*h = (g - b) / delta;
      else if (g == max)
	*h = 2.0 + (b - r) / delta;
      else if (b == max)
	*h = 4.0 + (r - g) / delta;

      *h = *h * 60.0;

      if (*h < 0.0)
	*h = *h + 360;
    }
}


void gdk_imlib_hsv_rotate(GdkImlibImage *im, gdouble hue, 
			  gdouble saturation, gdouble brightness,
			  gdouble alpha)
{
  int i,size = im->rgb_width*im->rgb_height;
  guchar *rgb = im->rgb_data, *alpha_data = im->alpha_data;
  GdkImlibColor transparency;
  gdouble r, g, b, h, s, v;
  gint a;

  gdk_imlib_get_image_shape(im, &transparency);

  for (i=0; i < size; i++) {
    /* Don't translate pixel if it is transparent. */
    if (rgb[0] == transparency.r && 
	rgb[1] == transparency.g &&
	rgb[2] == transparency.b)
      rgb += 3;
    else {
      /* Convert to hsv, do transformation, and convert back */
      rgb_to_hsv(rgb[0]/256.0, rgb[1]/256.0, rgb[2]/256.0, &h, &s, &v);
      h = fmod(h + hue + 360.0, 360.0); /* Note: if s or v are 0, h=-1 and
					   fmod doesn't do what I want, so
					   force h positive */
      s = s + saturation; if (s >= 1.0) s = 0.9999; else if (s < 0.0) s = 0.0;
      v = v + brightness; if (v >= 1.0) v = 0.9999; else if (v < 0.0) v = 0.0;
      hsv_to_rgb(h, s, v, &r, &g, &b);
      *rgb++ = (guchar)(r*256.0);
      *rgb++ = (guchar)(g*256.0);
      *rgb++ = (guchar)(b*256.0);
      if (alpha_data) {
	/* simply add the alpha translation */
	a = *alpha_data + (gint)rint(alpha*256.0);
	if (a >= 255) *alpha_data++ = 255; 
	else if (a <= 0) *alpha_data++ = 0;
	else *alpha_data++ = a;
      }
    }
  }

  gdk_imlib_changed_image(im);
}

/* Single step rotation/scaling/mirroring of an image.  The rotation
 * is through an arbitrary angle expressed in radians. The scaling 
 * applies to both axes equally (no matter what borders are defined). 
 * The mirroring is vertical only.  The portion of the new image 
 * outside the orignal image will be made transparent (all four corners). 
 * Note that this cannot happen in place since the new image will be of 
 * a different size from the old. Note also that rotating and unrotating
 * an image will not yield the same image. For one thing, it will be
 * bigger since I do not crop transparent borders (create another
 * function for this purpose if you need it since there is no advantage
 * to doing it in place).  For another thing, the truncation from
 * real coordinates to integer coordinates may cause some pixels to shift.
 */

#undef USE_BILINEAR
#undef MIRROR_ORIGINAL

GdkImlibImage *gdk_imlib_transform_image(GdkImlibImage *im, 
		gfloat scale, gfloat angle, gboolean flip)
{
  /* Compute rotation matrix */
  const gfloat sin_th = sin(angle);
  const gfloat cos_th = cos(angle);
  gfloat x, y, x1, y1, x2, y2;
  gfloat icos, jcos, isin, jsin;
#ifdef USE_BILINEAR
  gfloat dx, dy, p1, p2, p3, p4;
  guchar *z1, *z2, *z3, *z4;
  guchar a1, a2, a3, a4;
#else
  guchar *src;
#endif
  int new_w, new_h, i, j, pt_x, pt_y;
  guchar *rgb_data, *alpha_data, *rgb_ptr, *alpha_ptr;
  GdkImlibColor transparency;
  guchar transparency_src[3];
#ifdef HSV_INTERP
  double h1,s1,v1,h2,s2,v2,h3,s3,v3,h4,s4,v4,r,g,b;
#endif


  /* Compute new image size by scaling and rotating two of the corners
   * and picking the maximum dimension. Since we are rotating about the 
   * center, all the corners are going to be at (+- a*w/2, +- a*h/2).
   * We need the absolute value since the new corner may be anywhere 
   * in the circle.  We need to multiply the result by two since the 
   * coordinates are ot the radius, not the diameter. */
  x  = 0.5*im->rgb_width*scale;
  y  = 0.5*im->rgb_height*scale; 
  x1 = x*cos_th + y*sin_th,  y1 = y*cos_th - x*sin_th;
  x2 = -x*cos_th + y*sin_th, y2 = y*cos_th + x*sin_th;
  new_w = (int)(2.0*MAX(fabs(x1),fabs(x2)))+1;  
  new_h = (int)(2.0*MAX(fabs(y1),fabs(y2)))+1;
  TRACE("new (w,h) = (%d,%d)", new_w, new_h);

#if 1
  /* Simple optimization: if there is no rotation or scaling, simply
     clone and possibly flip the image. */
  if ((int)(2.0*x1) == im->rgb_width && (int)(2.0*y1) == im->rgb_height) {
    im = gdk_imlib_clone_image(im);
    if (flip) gdk_imlib_flip_image_horizontal(im);
    return im;
  }
#else
  WARNING! THIS SECTION IS INCOMPLETE;
  /* Better optimization: if the scaled dimensions are equal to the original
     dimensions (up to a pixel width) and the new sticker will fit into
     the old sticker, then we can produce the new image in place by
     some combination of flips. We could do even better by placing all
     the rotations and flips in the same function (i.e., the unused
     parameter to rotate, plus an encoding for flips). */
  if (can do it in place) {
    im = gdk_imlib_clone_image(im);
    if (flip) {
      if (0 degrees) gdk_imlib_flip_image_vertical(im);
      else if (90 degrees) { 
	gdk_imlib_flip_image_vertical(im);
	gdk_imlib_flip_image_horizontal(im);
	gdk_imlib_rotate_image(im, 1);
      }
      else if (180 degrees) gdk_imlib_flip_image_horizontal(im);
      else if (270 degrees) gdk_imlib_rotate_image(im, 1);
    }
    else {
      if (0 degrees) ;
      else if (90 degrees) {
	gdk_imlib_rotate_image(im, 1);
	gdk_imlib_flip_image_horizontal(im);
      }
      else if (180 degrees) {
	gdk_imlib_flip_image_vertical(im);
	gdk_imlib_flip_image_horizontal(im);
      }
      else if (270 degrees) {
	gdk_imlib_flip_image_vertical(im);
	gdk_imlib_rotate_image(im, 1);
      }
    }
    return im;
  }
#endif

  /* Create new space for the new image */
  rgb_ptr = rgb_data = malloc(new_w*new_h*3);
  alpha_ptr = alpha_data = im->alpha_data?malloc(new_w*new_h):NULL;

  /* Create a transparent cell for the parts that weren't on the
     original image. Note that if there is no transparent color
     defined, white will be used. */
  gdk_imlib_get_image_shape(im, &transparency);
  transparency_src[0] = transparency.r & 0xFF;
  transparency_src[1] = transparency.g & 0xFF;
  transparency_src[2] = transparency.b & 0xFF;

  /* Step through the cells of the new image one by one, computing
   * the unrotating and unscaling the pixel to see where it comes
   * from in the original image. 
   *
   * To do this, translate the pixel coordinate origin from the top-left
   * to the center (i,j)->(i-w/2,j-h/2), multiply by the rotation 
   * matrix for negative theta:
   *     / cos th  sin th\
   *     \-sin th  cos th/
   * undo the scaling by mulitplying by 1/scale, and finally translate 
   * the pixel coordinate origin form the center to the top-left
   * of the original image (i'-w'/2,j-h'/2)->(i',j').
   * is by 1/scale, and we have to translate the resulting coordinates
   * back with the origin at the top left rather than the center of 
   * the image. The result of this is:
   *     i -> (i*cos th + j*sin th + a*w/2)/a
   *     j -> (j*cos th - i*sin th + a*h/2)/a
   * It is trivial to flip the results in across the vertical before
   * rotation by flipping the "untransformed" x component (for i->x)
   * [MIRROR_ORIGINAL], but it is more natural to mirror after the
   * rotation. We do this by traversing the scan lines in the reverse
   * direction, or equivalently, by negating (i*cos th, i*sin th). This
   * will make your image look like it is rotating in the reverse
   * direction.
   *
   * Note that since we are stepping through the pixels by 1, we can
   * simply start at the top left and add sin(th)/cos(th) to the appropriate
   * values each step of the way.  We needn't be terrible concerned with
   * round-off since we are adding similar-sized values every step of
   * the way, since there aren't that many steps, and since we are 
   * rounding to the nearest integer when we are done anyway. 
   */
  scale = 1.0/scale;
  for (j=0, jcos = -0.5*new_h*cos_th, jsin = -0.5*new_h*sin_th; 
       j < new_h; 
       j++, jcos+=cos_th, jsin+=sin_th) {
    for (i = 0, icos = -0.5*new_w*cos_th, isin = -0.5*new_w*sin_th; 
	 i < new_w; 
	 i++, icos+=cos_th, isin+=sin_th) {
      
      /* Transform (i,j) -> (x,y) */
#ifdef MIRROR_ORIGINAL
      x1 = (icos + jsin + x)*scale;
      y1 = (jcos - isin + y)*scale;
      if (flip) x1 = (gfloat)im->rgb_width - 1.0 - x1;
#else
      if (flip) {
	x1 = (-icos + jsin + x)*scale;
	y1 = ( jcos + isin + y)*scale;
      }
      else {
	x1 = (icos + jsin + x)*scale;
	y1 = (jcos - isin + y)*scale;
      }
#endif

#ifndef USE_BILINEAR /* use nearest pixel value */

      /* Find the nearest pixel */
      pt_x = rint(x1), pt_y = rint(y1);

      /* Use transparent pixel if source is not in image */
      if (pt_x<0 || pt_y<0 || pt_x>=im->rgb_width || pt_y>=im->rgb_height) {
	src = transparency_src;
	if (alpha_data) *alpha_ptr++ = 0;
      }
      else {
	src = &im->rgb_data[3*(pt_y * im->rgb_width + pt_x)];
        if (alpha_data) *alpha_ptr++ = im->alpha_data[pt_y*im->rgb_width+pt_x];
      }

      /* Update the image pixel */
      *rgb_ptr++ = *src++;
      *rgb_ptr++ = *src++;
      *rgb_ptr++ = *src++;

#else  /* USE_BILINEAR */
#define IN(i,j) ((i)>=0 && (i)<im->rgb_width && (j)>=0 && (j)<im->rgb_height)
#define RGB(i,j) (im->rgb_data + 3*((j)*im->rgb_width + (i)))
#define ISTRANS(z) (z[0] == transparency_src[0] && z[1]==transparency_src[1] \
		    && z[2] == transparency_src[2])
#define ALPHA(i,j) (alpha_ptr?im->alpha_data[(j)*im->rgb_width + (i)]:255)
#define TRANS(a,b,c,d) (p1*(float)a+p2*(float)b+p3*(float)c+p4*(float)d)

      /* Find the surrounding in square */
      pt_x = floor(x1); pt_y = floor(y1);

      /* Find surrounding peaks */
      
      if (IN(pt_x,pt_y))     z1=RGB(pt_x,pt_y),     a1=ALPHA(pt_x,pt_y);
      else                   z1=transparency_src,   a1=0;
      if (IN(pt_x+1,pt_y))   z2=RGB(pt_x+1,pt_y),   a2=ALPHA(pt_x+1,pt_y);
      else                   z2=transparency_src,   a2=0;
      if (IN(pt_x,pt_y+1))   z3=RGB(pt_x,pt_y+1),   a3=ALPHA(pt_x,pt_y+1);
      else                   z3=transparency_src,   a3=0;
      if (IN(pt_x+1,pt_y+1)) z4=RGB(pt_x+1,pt_y+1), a4=ALPHA(pt_x+1,pt_y+1);
      else                   z4=transparency_src,   a4=0;

      /* Compute the transform */
      dx = x1 - (gfloat)pt_x; dy = y1 - (gfloat)pt_y;
      p1 = (1.0-dx)*(1.0-dy); p2 = dx*(1.0-dy); p3 = (1.0-dx)*dy; p4 = dx*dy;
      if (alpha_ptr) {
	*alpha_ptr++ = TRANS(a1, a2, a3, a4);
        *rgb_ptr++ = TRANS(*z1++,*z2++,*z3++,*z4++);
        *rgb_ptr++ = TRANS(*z1++,*z2++,*z3++,*z4++);
        *rgb_ptr++ = TRANS(*z1++,*z2++,*z3++,*z4++);
      }
      else if (ISTRANS(z1) || ISTRANS(z2) || ISTRANS(z3) || ISTRANS(z4)) {
	*rgb_ptr++ = transparency.r;
	*rgb_ptr++ = transparency.g;
	*rgb_ptr++ = transparency.b;
      }
      else {
	*rgb_ptr++ = TRANS(*z1++,*z2++,*z3++,*z4++);
	*rgb_ptr++ = TRANS(*z1++,*z2++,*z3++,*z4++);
	*rgb_ptr++ = TRANS(*z1++,*z2++,*z3++,*z4++);
      }

#endif /* USE_BILINEAR */
    }
  }

  TRACE("before with (w,h) = (%d,%d)", new_w, new_h);
  im = gdk_imlib_create_image_from_data(rgb_data, alpha_data, new_w, new_h);
  gdk_imlib_set_image_shape(im, &transparency);
#if 0
    if (flip) gdk_imlib_flip_image_horizontal(im);
#endif
  free(rgb_data); if (alpha_data) free(alpha_data);
  return im;
}



/* ===================================================================== */
/* Pointer confinement */
static GList *confine_list=NULL;

void confine_add(GtkWidget *widget)
{
  confine_list = g_list_prepend(confine_list, widget->window);
}

void confine_remove(GtkWidget *widget)
{
  confine_list = g_list_remove(confine_list, widget->window);
}

void confine_pointer(GtkWidget *widget)
{
  GtkWidget *root;
  GdkWindow *pointer_window;
  gint err;

  /* Locate root of widget tree, and the location of the pointer */
  root = gtk_widget_get_toplevel(widget);
  pointer_window = gdk_window_get_pointer(root->window,NULL,NULL,NULL);
  if (pointer_window == NULL) pointer_window = root->window;

  /* Release any current grabs before grabbing again */
  if (gdk_pointer_is_grabbed()) gdk_pointer_ungrab(GDK_CURRENT_TIME);

  /* If the pointer is in the canvas, we want to avoid a redraw, so let
   * the canvas keep it. It will give back the pointer when it leaves. */
  if (g_list_find(confine_list, pointer_window) && widget == root) {
    err = gdk_pointer_grab (pointer_window, TRUE, 0,
			    widget->window, NULL, GDK_CURRENT_TIME);
    if (err) TRACE("canvas confine_cursor: pointer grab=%d", err); 
  }
  else {
    err = gdk_pointer_grab (widget->window, TRUE, 0, 
			    widget->window, NULL, GDK_CURRENT_TIME);
    if (err) TRACE("widget confine_cursor: pointer grab=%d", err); 
    
    /* Force entry into the window containing the pointer */
    if (pointer_window!=widget->window && widget == root) {
      gdk_window_hide(pointer_window);
      gdk_window_show(pointer_window);
    }
  }
}


#ifndef HAVE_GRAB_IMAGE
/*
-------------------------------------------------------------------------------
                                I   M   L   I   B
                                      1.9.4
-------------------------------------------------------------------------------

This software is Copyright (C) 1998 By The Rasterman (Carsten Haitzler). I
accept no responsability for anythign this software may or may not do to
your system - you use it completely at your own risk. This software comes
under the LGPL and GPL licences. The library itself is LGPL (all software in
Imlib and gdk_imlib directories) see the COPYING and COPYING.LIB files for
full legal details.

-------------------------------------------------------------------------------
*/
#include <gdk/gdkx.h>
#include <gdk_imlib_private.h>
#ifdef HAVE_SHM
#include <sys/shm.h>
#include <X11/extensions/XShm.h>
#endif
#if 0
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/Xos.h>
#include <X11/extensions/shape.h>
#include <X11/cursorfont.h>
#endif

static char         x_error = 0;
static void
__handle_x_error(Display * d, XErrorEvent * ev)
{
  d = NULL;
  ev = NULL;
  x_error = 1;
}

GdkImlibImage      *
gdk_imlib_create_image_from_drawable(GdkWindow * gwin, GdkBitmap * gmask,
				     int x, int y, int width, int height)
{
  unsigned char      *data = NULL, *ptr, r, g, b;
  unsigned long       pixel;
  int                 i, xx, yy, w, h, inx, iny, clipx, clipy, rx, ry;
  XImage             *xim;

#ifdef HAVE_SHM
  XShmSegmentInfo     shminfo;

#endif
  XWindowAttributes   xatt, ratt;
  Colormap            cmap;
  static char         shm_checked = 0, shm = 1;
  XErrorHandler       erh = NULL;
  Window              chld;
  char                is_pixmap = 0;
  Visual             *vis;
  GdkImlibImage      *im;
  GdkImlibColor       ctab[256];
  Display            *disp;
  Window              win;
  Pixmap              mask;

  inx = 0;
  iny = 0;
  w = width;
  h = height;
  vis = id->x.visual;
  disp = id->x.disp;
  win = GDK_WINDOW_XWINDOW(gwin);
  if (gmask)
    mask = GDK_WINDOW_XWINDOW(gmask);
  else
    mask = 0;
  XGrabServer(disp);
  erh = XSetErrorHandler((XErrorHandler) __handle_x_error);
  x_error = 0;
  XGetWindowAttributes(disp, win, &xatt);
  XFlush(disp);
  if (x_error)
    {
      x_error = 0;
      is_pixmap = 1;
      XGetGeometry(disp, win, &chld, &rx, &rx,
		   (unsigned int *)&xatt.width, (unsigned int *)&xatt.height,
		   (unsigned int *)&rx, (unsigned int *)&xatt.depth);
      XFlush(disp);
      if (x_error)
	{
	  XUngrabServer(disp);
	  XFlush(disp);
	  XSetErrorHandler((XErrorHandler) erh);
	  return NULL;
	}
    }
  XSetErrorHandler((XErrorHandler) erh);
  if (!is_pixmap)
    {
      XGetWindowAttributes(disp, xatt.root, &ratt);
      XTranslateCoordinates(disp, win, xatt.root, 0, 0, &rx, &ry, &chld);
      if ((xatt.map_state != IsViewable) &&
	  (xatt.backing_store == NotUseful))
	{
	  XUngrabServer(disp);
	  XFlush(disp);
	  return NULL;
	}
    }
  clipx = 0;
  clipy = 0;

  x = x - inx;
  y = y - iny;

  width = xatt.width - x;
  height = xatt.height - y;
  if (width > w)
    width = w;
  if (height > h)
    height = h;

  if (!is_pixmap)
    {
      if ((rx + x + width) > ratt.width)
	width = ratt.width - (rx + x);
      if ((ry + y + height) > ratt.height)
	height = ratt.height - (ry + y);
    }
  if (x < 0)
    {
      clipx = -x;
      width += x;
      x = 0;
    }
  if (y < 0)
    {
      clipy = -y;
      height += y;
      y = 0;
    }
  if (!is_pixmap)
    {
      if ((rx + x) < 0)
	{
	  clipx -= (rx + x);
	  width += (rx + x);
	  x = -rx;
	}
      if ((ry + y) < 0)
	{
	  clipy -= (ry + y);
	  height += (ry + y);
	  y = -ry;
	}
    }
  if ((width <= 0) || (height <= 0))
    {
      XUngrabServer(disp);
      XSync(disp, False);
      return NULL;
    }
#ifdef HAVE_SHM
  if (shm)
    {
      if (!shm_checked)
	{
	  erh = XSetErrorHandler((XErrorHandler) __handle_x_error);
	}
      xim = XShmCreateImage(disp, vis, xatt.depth, ZPixmap, NULL,
			    &shminfo, width, height);
      if (!shm_checked)
	{
	  XSync(disp, False);
	  if (x_error)
	    {
	      shm = 0;
	      XDestroyImage(xim);
	      xim = XGetImage(disp, win, x, y, width, height, 0xffffffff, ZPixmap);
	      XSetErrorHandler((XErrorHandler) erh);
	      shm_checked = 1;
	    }
	  else
	    {
	      shminfo.shmid = shmget(IPC_PRIVATE, xim->bytes_per_line *
				     xim->height, IPC_CREAT | 0666);
	      if (shminfo.shmid < 0)
		{
		  shm = 0;
		  XDestroyImage(xim);
		  xim = XGetImage(disp, win, x, y, width, height, 0xffffffff, ZPixmap);
		  XSetErrorHandler((XErrorHandler) erh);
		  shm_checked = 1;
		}
	      else
		{
		  shminfo.shmaddr = xim->data = shmat(shminfo.shmid, 0, 0);
		  if (shminfo.shmaddr == (char *)-1)
		    {
		      shm = 0;
		      XDestroyImage(xim);
		      shmctl(shminfo.shmid, IPC_RMID, 0);
		      xim = XGetImage(disp, win, x, y, width, height, 0xffffffff, ZPixmap);
		      XSetErrorHandler((XErrorHandler) erh);
		      shm_checked = 1;
		    }
		  else
		    {
		      shminfo.readOnly = False;
		      XShmAttach(disp, &shminfo);
		    }
		}
	    }
	}
      else
	{
	  shminfo.shmid = shmget(IPC_PRIVATE, xim->bytes_per_line *
				 xim->height, IPC_CREAT | 0666);
	  if (shminfo.shmid < 0)
	    {
	      shm = 0;
	      XDestroyImage(xim);
	      xim = XGetImage(disp, win, x, y, width, height, 0xffffffff, ZPixmap);
	      XSetErrorHandler((XErrorHandler) erh);
	      shm_checked = 1;
	    }
	  else
	    {
	      shminfo.shmaddr = xim->data = shmat(shminfo.shmid, 0, 0);
	      if (shminfo.shmaddr == (char *)-1)
		{
		  shm = 0;
		  XDestroyImage(xim);
		  shmctl(shminfo.shmid, IPC_RMID, 0);
		  xim = XGetImage(disp, win, x, y, width, height, 0xffffffff, ZPixmap);
		  XSetErrorHandler((XErrorHandler) erh);
		  shm_checked = 1;
		}
	      else
		{
		  shminfo.readOnly = False;
		  XShmAttach(disp, &shminfo);
		}
	    }
	}
      if (!shm_checked)
	{
	  XSync(disp, False);
	  if (x_error)
	    {
	      shm = 0;
	      XDestroyImage(xim);
	      xim = XGetImage(disp, win, x, y, width, height, 0xffffffff, ZPixmap);
	      shm_checked = 1;
	    }
	  XSetErrorHandler((XErrorHandler) erh);
	  shm_checked = 1;
	}
    }
  else
#endif /* HAVE_SHM */
    xim = XGetImage(disp, win, x, y, width, height, 0xffffffff, ZPixmap);
#ifdef HAVE_SHM
  if (shm)
    XShmGetImage(disp, win, xim, x, y, 0xffffffff);
#endif
  XUngrabServer(disp);
  XFlush(disp);

  if (xatt.depth == 1)
    {
      ctab[0].r = 255;
      ctab[0].g = 255;
      ctab[0].b = 255;
      ctab[1].r = 0;
      ctab[1].g = 0;
      ctab[1].b = 0;
    }
  if (xatt.depth <= 8)
    {
      XColor              cols[256];

      if (!is_pixmap)
	{
	  cmap = xatt.colormap;
	  if (cmap == None)
	    cmap = id->x.root_cmap;
	}
      else
	cmap = id->x.root_cmap;

      for (i = 0; i < (1 << id->x.depth); i++)
	{
	  cols[i].pixel = i;
	  cols[i].flags = DoRed | DoGreen | DoBlue;
	}
      XQueryColors(disp, cmap, cols, 1 << id->x.depth);
      for (i = 0; i < (1 << id->x.depth); i++)
	{
	  ctab[i].r = cols[i].red >> 8;
	  ctab[i].g = cols[i].green >> 8;
	  ctab[i].b = cols[i].blue >> 8;
	  ctab[i].pixel = cols[i].pixel;
	}
    }
  data = malloc(width * height * 3);
  if (data)
    {
      ptr = data;
      switch (xatt.depth)
	{
	case 0:
	case 1:
	case 2:
	case 3:
	case 4:
	case 5:
	case 6:
	case 7:
	case 8:
	  for (yy = 0; yy < height; yy++)
	    {
	      for (xx = 0; xx < width; xx++)
		{
		  pixel = XGetPixel(xim, xx, yy);
		  r = ctab[pixel & 0xff].r;
		  g = ctab[pixel & 0xff].g;
		  b = ctab[pixel & 0xff].b;
		  *ptr++ = r;
		  *ptr++ = g;
		  *ptr++ = b;
		}
	    }
	  break;
	case 16:
	case 15:
	  if (id->x.render_depth == 16)
	    {
	      for (yy = 0; yy < height; yy++)
		{
		  for (xx = 0; xx < width; xx++)
		    {
		      pixel = XGetPixel(xim, xx, yy);
		      r = (pixel >> 8) & 0xf8;
		      g = (pixel >> 3) & 0xfc;
		      b = (pixel << 3) & 0xf8;
		      *ptr++ = r;
		      *ptr++ = g;
		      *ptr++ = b;
		    }
		}
	    }
	  else
	    {
	      for (yy = 0; yy < height; yy++)
		{
		  for (xx = 0; xx < width; xx++)
		    {
		      pixel = XGetPixel(xim, xx, yy);
		      r = (pixel >> 7) & 0xf8;
		      g = (pixel >> 3) & 0xf8;
		      b = (pixel << 3) & 0xf8;
		      *ptr++ = r;
		      *ptr++ = g;
		      *ptr++ = b;
		    }
		}
	    }
	  break;
	case 24:
	case 32:
	  for (yy = 0; yy < height; yy++)
	    {
	      for (xx = 0; xx < width; xx++)
		{
		  pixel = XGetPixel(xim, xx, yy);
		  r = (pixel >> 16) & 0xff;
		  g = (pixel >> 8) & 0xff;
		  b = pixel & 0xff;
		  *ptr++ = r;
		  *ptr++ = g;
		  *ptr++ = b;
		}
	    }
	  break;
	default:
	  for (yy = 0; yy < height; yy++)
	    {
	      for (xx = 0; xx < width; xx++)
		{
		  r = rand();
		  g = rand();
		  b = rand();
		  *ptr++ = r;
		  *ptr++ = g;
		  *ptr++ = b;
		}
	    }
	  break;
	}
    }

#ifdef HAVE_SHM
  if (shm)
    {
      XSync(disp, False);
      XShmDetach(disp, &shminfo);
      shmdt(shminfo.shmaddr);
      shmctl(shminfo.shmid, IPC_RMID, 0);
    }
#endif
  XDestroyImage(xim);

  if (data)
    {
      im = gdk_imlib_create_image_from_data(data, NULL, width, height);
      free(data);
      return im;
    }
  else
    return NULL;
}

#endif /* NEED_GRAB_IMAGE */

