/* rc.c
 *
 * Copyright 1999, Paul Kienzle (pkienzle@cs.indiana.edu)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * * July 25, 1999 - PAK
 *   Warn if saving a configuration containing relative paths
 * * May 13, 1999 - PAK
 *   Added AUTHOR/EMAIL text to configure window
 * * Initial revision
 */

#define VERBOSE 0
#include <stdlib.h>
#include <ctype.h>
#include "stickers.h"
#include "misc.h"

static gboolean use_tooltips = TRUE;
static GtkTooltips *tooltips = NULL;

void config_tip(GtkWidget *widget, gchar *text)
{
  if (tooltips == NULL) tooltips = GTK_TOOLTIPS(gtk_tooltips_new());
  gtk_tooltips_set_tip(tooltips, widget, text, NULL);
}

gchar *config_signature()
{
  if (config.signature[0]) {
    gchar base[256];
    time_t time_whole;
    struct tm *time_parts;

    time(&time_whole);
    time_parts = localtime(&time_whole);
#ifdef HAVE_STRFTIME
    strftime(base, sizeof(base), config.signature, time_parts);
#else
    snprintf(base, sizeof(base), "%s %d-%d-%d\n", config.signature,
	     time_parts.year, time_parts.month, time_parts.day);
#endif
    return g_strdup(base);
  }
  else return g_strdup("");
}


/* Bitmap data of LARGE RED CURSOR */
static const gchar *bigcursor_bits[] = {
"1111111111111111111111111               ",
"111111111111111111111111                ",
"1100000000000000000111                  ",
"11000000000000000111                    ",
"110000000000000111                      ",
"1100000000000111                        ",
"11000000000111                          ",
"110000000011                            ",
"1100000000011                           ",
"11000000000011                          ",
"110000110000011                         ",
"1100001110000011                        ",
"1100011 110000011                       ",
"1100011  110000011                      ",
"110011    110000011                     ",
"110011     110000011                    ",
"11011       110000011                   ",
"11011        110000011                  ",
"1111          110000011                 ",
"1111           110000011                ",
"111             110000011               ",
"111              110000011              ",
"11                110000011             ",
"1                  110000011            ",
"                    110000011           ",
"                     110000011          ",
"                      110000011         ",
"                       110000011        ",
"                        110000011       ",
"                         110000011      ",
"                          110000011     ",
"                           110000011    ",
"                            110000011   ",
"                             110000011  ",
"                              110000011 ",
"                               110000011",
"                                11000011",
"                                 1100011",
"                                  111111",
"                                   11111"};
static const gchar *tinycursor_bits[] = { 
  " 1 ", 
  "101",
  " 1 "};

void config_default()
{
  GdkColor fg, bg;

  TRACE("default config");
  /* Define a big ugly cursor so kids can find it easily. */
  gdk_color_parse("rgb:0000/0000/0000",&fg);
  gdk_color_parse("rgb:FFFF/3FFF/0000",&bg);
  config.big_cursor = 
    gdk_cursor_new_from_data(bigcursor_bits, 40, 40, &fg, &bg, 0, 0);
  config.tiny_cursor =
    gdk_cursor_new_from_data(tinycursor_bits, 3, 3, &fg, &bg, 1, 1);

  /* Assign defaults */
  config.work_dir = g_strdup("./");
  config.sticker_dir = g_strdup(DATADIR);
  config.icon_dir = g_strdup(DATADIR);

  config.signature = g_strdup("");
  config.sigfont = g_strdup("");
  config.printfile = g_strdup("");
  config.printcmd = g_strdup("lp %s");
  config.button_size = 5;
  config.scene_size = 10;
  config.max_history = 30;
  config.max_saved = config.max_history;
  config.paint = FALSE;
  config.confine = FALSE;
  config.maximize = FALSE;
  config.print = FALSE;
  config.config = TRUE;
  config.use_big_cursor = TRUE;
  config.use_tiny_cursor = TRUE;
  config.initial_delay = 300; /* 300 ms between click and start of rotation */
  config.delay = 60;          /* 50 ms between steps of rotation */
}

static void getstring(FILE *file, gchar *string[])
{
  char buffer[256];
  int len;

  fgets (buffer, sizeof(buffer)-1, file);
  buffer[sizeof(buffer)-1] = '\0';
  for (len = strlen(buffer); len && isspace(buffer[len-1]); len--) ;
  buffer[len] = '\0';
  for (len = 0; isspace(buffer[len]); len++) ;
  g_free(*string);
  *string = g_strdup(buffer+len);
}

static void fix_dir (gchar **dir)
{
  int len = strlen(*dir);

  if (len && (*dir)[len-1] != '/') {
    gchar *new = g_strconcat(*dir, "/", NULL);
    g_free(*dir);
    *dir = new;
  }
}

void config_load(gchar *filename)
{
  gchar *home, *rcfile;
  gchar token[256];
  gint v;
  FILE *file;

  TRACE("enter");

  /* Assign the default configuration */
  config_default();

  /* look for resource file in $(HOME)/.stickers */
  home = getenv("HOME");
  if (home) {
    rcfile = g_strconcat(home, "/.stickers", NULL);
    file = fopen(rcfile, "r");
  }
  else file = NULL;

  /* load resource file */
  if (file != NULL) {
    config.paint = config.confine = config.maximize = use_tooltips =
      config.use_big_cursor = config.use_tiny_cursor = FALSE;
    v = fscanf(file, " @ %255s", token);
    TRACE("config token=%s", token);
    while (v == 1) {
      if (strcmp(token, "button_size") == 0)
	{ TRACE("scanning for button_size");
	v = fscanf(file, " %d ", &config.button_size);
	TRACE("found %d", config.button_size);
	}
      else if (strcmp(token, "thumb_size") == 0)
	v = fscanf(file, " %d ", &config.scene_size);
      else if (strcmp(token, "save_size") == 0)
	v = fscanf(file, " %d ", &config.max_saved);
      else if (strcmp(token, "undo_size") == 0)
	v = fscanf(file, " %d ", &config.max_history);
      else if (strcmp(token, "rotate_delay") == 0)
	v = fscanf(file, " %d ", &config.initial_delay);
      else if (strcmp(token, "rotate_rate") == 0)
	v = fscanf(file, " %d ", &config.delay);
      else if (strcmp(token, "paint") == 0) config.paint = TRUE;
      else if (strcmp(token, "print") == 0) config.print = TRUE;
      else if (strcmp(token, "config") == 0) config.config = TRUE;
      else if (strcmp(token, "confine") == 0) config.confine = TRUE;
      else if (strcmp(token, "maximize") == 0) config.maximize = TRUE;
      else if (strcmp(token, "bigcursor") == 0) config.use_big_cursor = TRUE;
      else if (strcmp(token, "tinycursor") == 0) config.use_tiny_cursor = TRUE;
      else if (strcmp(token, "tooltips") == 0) use_tooltips = TRUE;
      else if (strcmp(token, "sticker_dir") == 0) 
	getstring(file, &config.sticker_dir);
      else if (strcmp(token, "icon_dir") == 0) 
	getstring(file, &config.icon_dir);
      else if (strcmp(token, "work_dir") == 0) 
	getstring(file, &config.work_dir);
      else if (strcmp(token, "printfile") == 0) 
	getstring(file, &config.printfile);
      else if (strcmp(token, "printfile") == 0) 
	getstring(file, &config.printcmd);
      else if (strcmp(token, "signature") == 0) 
	getstring(file, &config.signature);
      else if (strcmp(token, "sigfont") == 0) 
	getstring(file, &config.sigfont);
      else fgets(token, sizeof(token), file);
      v = fscanf(file, " @ %255s", token);
    }
    fclose(file);
  }

  /* Finally, use the specified start directory */
  if (filename) {
    g_free(config.sticker_dir);
    config.sticker_dir = g_strdup(filename);
    g_free(config.icon_dir);
    config.icon_dir = g_strdup(filename);
  }

  /* Fixup missing trailing '/' on directories */
  fix_dir(&config.sticker_dir);
  fix_dir(&config.icon_dir);
  fix_dir(&config.work_dir);

  /* Turn on tooltips in the rest of stickers, if desired */
  if (tooltips == NULL) tooltips = GTK_TOOLTIPS(gtk_tooltips_new());
  if (use_tooltips) gtk_tooltips_enable(tooltips);
  else gtk_tooltips_disable(tooltips);
}


void config_save(void)
{
  gchar *home, *rcfile;
  FILE *file;

  home = getenv("HOME");
  if (home == NULL) {
    printf(_("$HOME not defined?!.\n"));
    return;
  }

  if (config.sticker_dir[0] != '/'
      || config.icon_dir[0] != '/'
      || (config.work_dir[0] != '/' && !strcmp(config.work_dir, "./"))) {

    GtkWidget *box, *button;

    box=gtk_dialog_new();
    gtk_signal_connect (GTK_OBJECT (box), "destroy",
			GTK_SIGNAL_FUNC(gtk_widget_destroyed),
			&box);
    gtk_window_set_title (GTK_WINDOW (box), _("Sticker Warning"));

    gtk_container_border_width (GTK_CONTAINER (GTK_DIALOG(box)->vbox), 10);
    gtk_box_pack_start (GTK_BOX( GTK_DIALOG(box)->vbox),
			gtk_label_new(_("You are using relative paths to the configured directories.\nYou probably don't want that, but I'll save them anyway.")),
			FALSE, FALSE, 0);

    button = gtk_button_new_with_label ("OK");
    gtk_signal_connect_object(GTK_OBJECT(button), "clicked",
			      GTK_SIGNAL_FUNC(gtk_widget_destroy),
			      GTK_OBJECT(box));
    GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
    gtk_box_pack_start (GTK_BOX (GTK_DIALOG (box)->action_area), 
			button, TRUE, TRUE, 0);
    gtk_widget_grab_default (button);
    gtk_widget_show_all (box);
  }
  rcfile = g_strconcat(home, "/.stickers", NULL);
  file = fopen(rcfile, "w");
  if (file == NULL) {
    printf(_("Could not open %s for writing.\n"), rcfile);
    return;
  }

  fprintf(file, "@button_size %d\n", config.button_size);
  fprintf(file, "@thumb_size %d\n", config.scene_size);
  fprintf(file, "@save_size %d\n", config.max_saved);
  fprintf(file, "@undo_size %d\n", config.max_history);
  fprintf(file, "@rotate_delay %d\n", config.initial_delay);
  fprintf(file, "@rotate_rate %d\n", config.delay);
  if (config.paint) fprintf(file, "@paint\n");
  if (config.print) fprintf(file, "@print\n");
  if (config.config) fprintf(file, "@config\n");
  if (config.confine) fprintf(file, "@confine\n");
  if (config.maximize) fprintf(file, "@maximize\n");
  if (config.use_big_cursor) fprintf(file, "@bigcursor\n");
  if (config.use_tiny_cursor) fprintf(file, "@tinycursor\n");
  if (use_tooltips) fprintf(file, "@tooltips\n");
  fprintf(file, "@sticker_dir %s\n", config.sticker_dir);
  fprintf(file, "@icon_dir %s\n", config.icon_dir);
  fprintf(file, "@work_dir %s\n", config.work_dir);
  fprintf(file, "@printfile %s\n", config.printfile);
  fprintf(file, "@printcmd %s\n", config.printcmd);
  fprintf(file, "@signature %s\n", config.signature);
  fprintf(file, "@sigfont %s\n", config.sigfont);
  fclose(file);
}

static void toggle_changed(GtkToggleButton *button, gboolean *state)
{
  *state = button->active;
}
static GtkWidget* 
add_toggle(GtkWidget *box, gchar *label, gboolean *state, gchar *tip)
{
  GtkWidget *button;

  button = gtk_check_button_new_with_label (label);
  gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON(button), *state);
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
		      GTK_SIGNAL_FUNC(toggle_changed),
		      state);
  gtk_box_pack_start (GTK_BOX (box), button, TRUE, TRUE, 0);
  if (tip) config_tip(button, tip);
  return button;
}

static void directory_changed(GtkEntry *entry, gchar *value[])
{
  gchar *text;

  g_free(*value);
  text = gtk_entry_get_text(entry);
  if (text[strlen(text)-1] != '/')
    *value = g_strconcat(text, "/", NULL);
  else
    *value = g_strdup(text);
}
static GtkWidget* 
add_directory(GtkWidget *box, gchar *label, gchar *value[], gchar *tip)
{
  GtkWidget *box2, *entry;

  box2 = gtk_hbox_new(FALSE, 10);
  gtk_box_pack_start(GTK_BOX(box), box2, FALSE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(box2), gtk_label_new(label), FALSE, TRUE, 0);
  entry = gtk_entry_new();
  gtk_entry_set_text(GTK_ENTRY(entry), *value);
  gtk_entry_select_region(GTK_ENTRY(entry), 0, strlen(*value));
  gtk_signal_connect (GTK_OBJECT (entry), "changed",
		      GTK_SIGNAL_FUNC(directory_changed),
		      value);
  gtk_box_pack_end(GTK_BOX(box2), entry, TRUE, TRUE, 0);
  if (tip) config_tip (entry, tip);
  return entry;
}

static void text_changed(GtkEntry *entry, gchar *value[])
{
  gchar *text;

  g_free(*value);
  text = gtk_entry_get_text(entry);
  *value = g_strdup(text);
}
static GtkWidget* 
add_text(GtkWidget *box, gchar *label, gchar *value[], gchar *tip)
{
  GtkWidget *box2, *entry;

  box2 = gtk_hbox_new(FALSE, 10);
  gtk_box_pack_start(GTK_BOX(box), box2, FALSE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(box2), gtk_label_new(label), FALSE, TRUE, 0);
  entry = gtk_entry_new();
  gtk_entry_set_text(GTK_ENTRY(entry), *value);
  gtk_entry_select_region(GTK_ENTRY(entry), 0, strlen(*value));
  gtk_signal_connect (GTK_OBJECT (entry), "changed",
		      GTK_SIGNAL_FUNC(text_changed),
		      value);
  gtk_box_pack_end(GTK_BOX(box2), entry, TRUE, TRUE, 0);
  if (tip) config_tip (entry, tip);
  return entry;
}

static void spinner_changed(GtkAdjustment *adj, gint *value)
{
  *value = adj->value;
}
static GtkAdjustment* add_spinner(GtkWidget *box, gchar *label, 
			      gint *value, gint lower, gint upper, gchar *tip)
{
  GtkWidget *box2, *spinner;
  GtkAdjustment *adj;

  box2 = gtk_hbox_new(FALSE, 10);
  gtk_box_pack_start(GTK_BOX(box), box2, FALSE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(box2), gtk_label_new(label), FALSE, TRUE, 0);
  adj = (GtkAdjustment *) gtk_adjustment_new ((gfloat)*value, 
					      (gfloat)lower, 
					      (gfloat)upper, 
					      1.0, 5.0, 0.0);
  spinner = gtk_spin_button_new (adj, 0, 0);
  gtk_box_pack_end (GTK_BOX (box2), spinner, TRUE, TRUE, 0);
  gtk_signal_connect (GTK_OBJECT (adj), "value_changed",
		      GTK_SIGNAL_FUNC(spinner_changed),
		      value);
  if (tip) config_tip (spinner, tip);
  return adj;
}

static void font_changed(GtkWidget *button, GtkWidget *wptrs[])
{
  /* User clicked OK in the font selection dialog box wptrs[0], so get the
   * new font name, show it to the user in the display box wptrs[1] and
   * destroy the font selection dialog box, which sets wptrs[0] to NULL. */

  gchar *s;

  TRACE("[%p,%p]",wptrs[0],wptrs[1]);
  s = gtk_font_selection_dialog_get_font_name
    (GTK_FONT_SELECTION_DIALOG (wptrs[0]));
ONTRACE("s=%p", s);
  if (s) {
    gtk_entry_set_text(GTK_ENTRY(wptrs[1]), s);
    g_free(s);
  }
  else
    gtk_entry_set_text(GTK_ENTRY(wptrs[1]), "");
  gtk_widget_destroy (wptrs[0]);
}
static void font_request(GtkButton *button, GtkWidget *wptrs[])
{
  /* User clicked on the button to request the font selection dialog.
   * If the font selection dialog widget wptrs[0] exists, then
   * activate it. If it doesn't, then we need to create a new font
   * selection dialog and tell it to set wptrs[0] to NULL if it gets
   * destroyed. If the Cancel button is clicked, then destroy the
   * widget directly. If OK is clicked, update the font in the
   * display box then destroy the widget. */

  TRACE("[%p,%p]",wptrs[0],wptrs[1]);
  if (wptrs[0]) {
    gdk_window_raise(wptrs[0]->window);
  }
  else {
    GtkWidget *window;
    gchar *label; /* Grab the dialog title from the button label */
    gtk_label_get(GTK_LABEL(GTK_BIN(button)->child), &label);
    window = gtk_font_selection_dialog_new (label);
    wptrs[0] = window;
    gtk_window_set_position (GTK_WINDOW (window), GTK_WIN_POS_MOUSE);
    gtk_signal_connect (GTK_OBJECT (window), "destroy",
			GTK_SIGNAL_FUNC(gtk_widget_destroyed),
			&(wptrs[0]));
    gtk_signal_connect 
      (GTK_OBJECT (GTK_FONT_SELECTION_DIALOG (window)->ok_button),
       "clicked", GTK_SIGNAL_FUNC(font_changed), wptrs);
    gtk_signal_connect_object 
      (GTK_OBJECT (GTK_FONT_SELECTION_DIALOG (window)->cancel_button),
       "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT (window));
    gtk_widget_show (window);
  }
}
static void font_destroy(GtkButton *button, GtkWidget *wptrs[])
{
  /* When the configure window is destroyed, the font selection dialog
   * in wptrs[0] will not automatically be destroyed with it since it is not a
   * child of the configure window, but a whole toplevel dialog of
   * its own, so we have to manually destroy it. Don't forget to free
   * the wptrs! */
  TRACE("[%p,%p]",wptrs[0],wptrs[1]);
  if (wptrs[0]) gtk_widget_destroy(wptrs[0]);
  g_free(wptrs);
}
static GtkWidget* 
add_font(GtkWidget *box, gchar *label, gchar *value[], gchar *tip)
{
  /* Font selection widget */
  GtkWidget *box2, *button, *entry;
  GtkWidget **wptrs;

  TRACE("");
  box2 = gtk_hbox_new(FALSE, 10);
  button = gtk_button_new_with_label (label);
  entry = gtk_entry_new ();
  gtk_box_pack_start(GTK_BOX(box), box2, FALSE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(box2), button, FALSE, TRUE, 0);
  gtk_box_pack_end (GTK_BOX (box2), entry, TRUE, TRUE, 0);

  gtk_entry_set_text (GTK_ENTRY(entry), *value);
  gtk_entry_set_editable (GTK_ENTRY(entry), FALSE);
  gtk_signal_connect (GTK_OBJECT (entry), "changed",
		      GTK_SIGNAL_FUNC(text_changed),
		      value);

  wptrs = g_new (GtkWidget *, 2);
  wptrs[0] = NULL;
  wptrs[1] = entry;
  gtk_signal_connect (GTK_OBJECT(button), "clicked",
		      GTK_SIGNAL_FUNC (font_request),
		      wptrs);
  gtk_signal_connect (GTK_OBJECT(button), "destroy",
		      GTK_SIGNAL_FUNC (font_destroy),
		      wptrs);
  if (tip) config_tip (button, tip), config_tip (entry, tip);
  return entry;
}

static GtkWidget* add_label(GtkWidget *box, gchar *label, gchar value[])
{
  GtkWidget *box2, *widget;

  box2 = gtk_hbox_new(FALSE, 10);
  gtk_box_pack_start(GTK_BOX(box), box2, FALSE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(box2), gtk_label_new(label), FALSE, TRUE, 0);
  widget = gtk_label_new (value);
  gtk_box_pack_end (GTK_BOX (box2), widget, TRUE, TRUE, 0);

  return widget;
}

void config_close (GtkWidget *widget, GtkWidget **widget_pointer)
{
  if (widget_pointer) *widget_pointer = NULL;
  gtk_main_quit();
}

static void
update_sample_font(GtkWidget *fontentry, GtkWidget *label)
{
  GtkStyle *style;
  GdkFont *font;

  font = NULL;
  if (config.sigfont[0]) font = gdk_font_load(config.sigfont);
  if (font == NULL) font = gdk_font_ref(fontentry->style->font);

  /* Set new font style */
  style = gtk_style_new ();
  gdk_font_unref (style->font);
  style->font = font;
  gdk_font_ref (style->font);
  gtk_widget_set_style (label, style);
  gtk_style_unref(style);
}
    
static void
update_sample_text(GtkWidget *entry, GtkWidget *label)
{
  gchar *sig = config_signature();

  /* Set new label text */
  gtk_label_set_text(GTK_LABEL(label), sig);
  g_free(sig);
}

void config_edit()
{
  static GtkWidget *window = NULL;
  GtkWidget *box1, *box2, *scrolled_window;
  GtkWidget *button, *sig, *font, *sample, *tips;
  char version[256];

TRACE("config edit.");
  if (!window)
    {
      window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
      gtk_widget_set_name (window, "configure");
      gtk_widget_set_usize (window, 500, 400);
      gtk_window_set_position (GTK_WINDOW (window), GTK_WIN_POS_MOUSE);

      gtk_signal_connect (GTK_OBJECT (window), "destroy",
			  GTK_SIGNAL_FUNC(config_close),
			  &window);
      gtk_window_set_title (GTK_WINDOW (window), 
			    _("Sticker Configure"));
      gtk_container_border_width (GTK_CONTAINER (window), 0);


      box1 = gtk_vbox_new (FALSE, 0);
      gtk_container_add (GTK_CONTAINER (window), box1);
      snprintf(version, sizeof(version), 
	       _("Sticker Book Version %s"), VERSION);
      gtk_box_pack_start (GTK_BOX (box1), 
			  gtk_label_new(version), FALSE, TRUE, 0);
      gtk_box_pack_start (GTK_BOX (box1), 
	 gtk_label_new(_("Brought to you by Paul Kienzle and Tanya Riseman")),
			  FALSE, TRUE, 0);
      gtk_box_pack_start (GTK_BOX (box1), 
	 gtk_label_new("stickers@kienzle.powernet.co.uk"), FALSE, TRUE,0);
      gtk_box_pack_start (GTK_BOX (box1), gtk_hseparator_new(), FALSE, TRUE,0);

      scrolled_window = gtk_scrolled_window_new (NULL, NULL);
      gtk_container_border_width (GTK_CONTAINER (scrolled_window), 0);
      gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled_window),
				      GTK_POLICY_AUTOMATIC, 
				      GTK_POLICY_AUTOMATIC);
      gtk_box_pack_start (GTK_BOX (box1), scrolled_window, TRUE, TRUE, 0);
      

      box2 = gtk_vbox_new (FALSE, 0);
      gtk_container_border_width (GTK_CONTAINER (box2), 10);
      gtk_scrolled_window_add_with_viewport 
	(GTK_SCROLLED_WINDOW (scrolled_window), box2);


      add_toggle(box2, _("Paint mode"), &config.paint, _("\
If set, allow \"painting\" with the current sticker. Otherwise you must \
click once to paste each sticker."));
      add_toggle(box2, _("Confine pointer"), &config.confine, _("\
If set, confine the mouse pointer to within the window. This is useful for \
young children with poor mouse control since it keeps them from accidentally \
clicking on the border and resizing the window."));
      add_toggle(box2, _("Full screen"), &config.maximize, _("\
If set, open scenes as large as possible. The stickers scale \
automatically with the scene, so you don't really get more space, but \
it does make it easier to place stickers where you want them."));
      add_toggle(box2, _("Use big ugly pointer"), 
		 &config.use_big_cursor, _("\
If set, use a very large red pointer, which makes it easier for young eyes \
to know where the mouse is."));
      add_toggle(box2, _("Invisible pointer in canvas"), 
		 &config.use_tiny_cursor, _("\
If set, use a very small red pointer on the canvas. This makes it easier \
to see how your scene will look after the sticker is pasted since the \
large red mouse pointer doesn't get in the way."));
      tips = add_toggle(box2, _("Show tooltips"), &use_tooltips, _("\
If set, show tooltips for scene and sticker controls while editing the scene."));
      add_toggle(box2, _("Config control"), &config.config, _("\
If set, add a configure button to the scene controls.  You can always \
access the configure window with \"!\", even if the config control \
button is not available.  Note that the configure button will not \
appear/disappear when you return to the scene, but only when you select \
a scene."));

      add_toggle(box2, _("Print control"), &config.print, _("\
If set, add a print button to scene controls.  Need to set the working \
directory, the saved image file and the print command for printing \
to work.  Note that the print button will not appear/disappear when you \
return to the scene, but only when you start a new scene.  For the current \
scene, you can use \"#\" to print, or exit and re-enter the scene and \
use the print button."));
      add_text(box2, _("Saved image file (%s for scene name)"), 
	       &config.printfile, _("\
File in which to save a printable image of the working scene. Use %s in the \
name if you want a separate image for each scene. Use an extension of .pnm, \
.jpeg, .png, .tiff, .eim, .gif, .xpm or .bmp. The file is stored in the \
working directory. For example, %s.jpeg"));
      add_text(box2, _("Print command (%s for saved image file)"), 
	       &config.printcmd, _("\
Command to print the saved image. Use %s where you would put the filename \
in the print command. For example, lp -Pcolor %s."));
      add_directory(box2, _("Working directory"), &config.work_dir, _("\
Directory in which to save the working scenes and the printable images."));

      add_directory(box2, _("Sticker directory"), &config.sticker_dir, _("\
Directory containing the scene files and the stickers."));
      add_directory(box2, _("Icon directory"), &config.icon_dir, _("\
Directory containing the button surface images for the controls."));

      font = add_font(box2, _("Signature font"), &config.sigfont, _("\
Font used to render the signature. Click the button to change the font."));
      sig = add_text(box2, _("Signature (see man 3 strftime)"), 
		     &config.signature, _("\
String to put on the bottom right hand corner of the image. It is formatted \
using strftime, so you can specify date information."));
      sample = add_label(box2, _("Signature sample"), "");
      update_sample_font(font, sample); 
      update_sample_text(sig, sample);
      gtk_signal_connect (GTK_OBJECT (font), "changed",
			  GTK_SIGNAL_FUNC(update_sample_font),
			  sample);
      gtk_signal_connect (GTK_OBJECT (sig), "changed",
			  GTK_SIGNAL_FUNC(update_sample_text),
			  sample);

      add_spinner(box2, _("Control button size (% screen width)"), 
		  &config.button_size, 1, 6, _("\
Sets the size of the buttons and stickers in the scene editor. Use large \
control buttons for younger children. I like them large, too."));
      add_spinner(box2, _("Scene button size (% screen width)"), 
		  &config.scene_size, 2, 12, _("\
Sets the size of the scene buttons on the entry screen."));
      add_spinner(box2, _("Max. stickers saved (x100kB)"), 
		  &config.max_saved, 1, 1000000, _("\
Sets the number of megabytes of stickers to save when you exit a scene. The \
choice depends on the speed of your machine. If it is too large, reloading \
the scene will take a long time."));
      add_spinner(box2, _("Max. undo stack size (x100kB)"), 
		  &config.max_history, 1, 1000, _("\
Sets the number of megabytes of stickers you can \"unstick\" with the undo \
control. The choice depends on the the amount of memory on your machine."));
      add_spinner(box2, _("Delay between keypress and rotation (msec)"),
		  &config.initial_delay, 1, 1000, _("\
Sets the delay between the first operation and repeated operation for \
the sticker controls. Increase it if your toddler is ham-fisted. 300 ms \
works well."));
      add_spinner(box2, _("Rotation speed (msec per step)"),
		  &config.delay, 1, 1000, _("\
Sets the delay between updates for sticker controls. Increase the \
step delay for your ham-fisted toddler so that they have time to pull \
their finger off their key when the sticker is just right. Decrease it \
for fast animation. If the delay is too short, then the computer cannot \
keep up and weird things start to happen."));

      gtk_box_pack_start (GTK_BOX (box1), gtk_hseparator_new(), FALSE, TRUE,0);

      box2 = gtk_vbox_new (FALSE, 10);
      gtk_container_border_width (GTK_CONTAINER (box2), 10);
      gtk_box_pack_start (GTK_BOX (box1), box2, FALSE, TRUE, 0);
      gtk_widget_show (box2);

      button = gtk_button_new_with_label (_("Save settings"));
      gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
				 GTK_SIGNAL_FUNC(config_save),
				 NULL);
      gtk_box_pack_start (GTK_BOX (box2), button, TRUE, TRUE, 0);
      gtk_widget_show (button);
      config_tip(button, _("\
Write the configuration information to $HOME/.stickers so it will be set \
the next time. Note that the configuration changes will still take effect \
for this session even if they are not saved."));

      button = gtk_button_new_with_label (_("Continue"));
      gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
				 GTK_SIGNAL_FUNC(gtk_widget_destroy),
				 GTK_OBJECT (window));
      gtk_box_pack_start (GTK_BOX (box2), button, TRUE, TRUE, 0);
      GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
      gtk_widget_grab_default (button);
      gtk_widget_show (button);
      config_tip(button, _("Return to stickers with the new configuration."));

      gtk_widget_show_all (box1);
    }

  if (!GTK_WIDGET_VISIBLE (window)) gtk_widget_show (window);

  if (!use_tooltips) gtk_tooltips_enable(tooltips);
  gtk_main();
  if (!use_tooltips) gtk_tooltips_disable(tooltips);
}
