/* select.c
 *
 * Copyright 2000, Paul Kienzle (pkienzle@kienzle.powernet.co.uk)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * * Oct 27, 2000 - PAK
 *   Chop from main
 *   Restructure into base/working pairs rather base set/working set
 *   ! to load config screen
 *   Start config if scene directory is not available
 */

#include <gdk/gdkkeysyms.h>
#include <math.h>
#include <stdlib.h>

#define VERBOSE 0  /* Whether to print trace messages */
#define USE_TABS 1 /* Whether sticker sets show folder tabs or not */
#include "stickers.h"

/* == Select scene ================================================= */
static int scene_name_match(const struct dirent *entry)
{
  if (NAMLEN(entry) < 6) return FALSE;
  return (strncmp(entry->d_name + NAMLEN(entry) - 6, ".scene", 6) == 0);
}

static void select_scene_start(GtkWidget *widget, gchar *filename)
{
  GtkWidget *toplevel = gtk_widget_get_toplevel(widget);

  TRACE("select scene start with %s", filename);

  /* Start the new scene before removing the scene select window so that
     there is something to see on the screen. */
  /* XXX FIXME XXX  why doesn't the cursor change like it is supposed to? */
  set_pointer_wait (toplevel);
  if (!scene_load(filename)) {
    set_pointer (toplevel);
    return;
  }
  scene_new();
//  set_pointer (toplevel);

  TRACE("start_scene is destroying %p", toplevel);
  gtk_widget_destroy(toplevel);
}
static void select_scene_quit()
{
  TRACE("select scene quit");
  gtk_exit(0);
}
static void select_scene_delete(GtkWidget *widget, GdkEvent *event, 
				char *filename)
{
  g_free(filename);
}

static GtkWidget *
select_scene_button(char *directory, struct dirent *file, int size)
{
  GtkWidget *button;
  gchar *ext, *name, *path;
  gint len;

  len = NAMLEN(file);
  name = g_new(gchar, len+1);
  memcpy(name, file->d_name, len);
  name[len] = '\0';
  path = g_strconcat(directory, name, ".xpm", NULL);
  ext = strrchr(name,'.');
  *ext = '\0'; /* <blah>.scene -> <blah> */
  button = button_new_with_image(path, size, name);
  *ext = '.'; /* <blah> -> <blah>.scene */
  g_free(path);
  path = g_strconcat(directory, name, NULL); /* free'd on delete_event */
  g_free(name);
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
		      GTK_SIGNAL_FUNC (select_scene_start), path);
  gtk_signal_connect (GTK_OBJECT (button), "delete_event",
		      GTK_SIGNAL_FUNC (select_scene_delete), path);
  return button;
}

static void select_scene_config(GtkWidget * widget)
{
  /* XXX FIX ME XXX */
  /* This is rather brutal flow control: what I want to do is hide the
     scene select window while the configuration window is up and reload
     the available scenes when it is finished so that if the scene
     directory changes the new set of scenes will be shown.  I do this
     by destroying the select_scene window (which generated the
     current keystroke), raising the configuration window, and
     recreating the select_scene window when I'm done.  It is ugly,
     but it seems to work. What I should do is monitor for changes in
     directory, and if found, clear the table, resize and reload. */
  gtk_widget_destroy(gtk_widget_get_toplevel(widget));
  config_edit(); 
  select_scene();
}

/* Process scene select keystrokes. */
static int select_scene_keystroke (GtkWidget * widget, GdkEventKey * event)
{
  if (event->type == GDK_KEY_PRESS) 
    TRACE("Pressing key %d", event->keyval);
  else TRACE("Releasing key %d", event->keyval);

  if (event->type == GDK_KEY_PRESS) switch(event->keyval) {
  case GDK_Escape:
    select_scene_quit(); break;
  case GDK_exclam:
    select_scene_config(widget); break;
  }

#if 0 /* Don't want to eat keystrokes here since table handles arrow keys. */
  if (event->type == GDK_KEY_PRESS)
    gtk_signal_emit_stop_by_name(GTK_OBJECT(widget), "key_press_event");
  else if (event->type == GDK_KEY_RELEASE)
    gtk_signal_emit_stop_by_name(GTK_OBJECT(widget), "key_release_event");
#endif

  return FALSE;
}

/* XXX FIXME XXX
 * Why doesn't alphasort work?
 */
int dirsort(struct dirent ** f1, struct dirent ** f2)
{
  int ilen = NAMLEN(*f1);
  int jlen = NAMLEN(*f2);
  int cmp = strncmp((*f1)->d_name, (*f2)->d_name, 
		    ilen<jlen ? ilen : jlen);
  if (cmp == 0) {
    if (ilen < jlen) cmp = -1;
    else if (ilen > jlen) cmp = 1;
  }
  return cmp;
}

static GtkWidget *build_scene_table()
{
  static gchar *base_tip=NULL, *work_tip=NULL;
  GtkWidget *table, *entry, *base, *work;
  gint count1, count2, i, j, cmp;
  gint across, down, rows, cols, size;
  struct dirent **files1, **files2;

#define SPACING 5 /* Spacing between columns */

  if (base_tip == NULL) base_tip = _("Click to create a scene such as this.");
  if (work_tip == NULL) work_tip = _("Click to modify this scene.");

  /* find the scenes */
  count1 = scandir (config.sticker_dir, &files1, scene_name_match, dirsort);
  if (count1 <= 0) return NULL;

  /* find the working versions of the scenes */
  if (strcmp(config.sticker_dir, config.work_dir) != 0)
    count2 = scandir (config.work_dir, &files2, scene_name_match, dirsort);
  else
    count2 = 0;
  
  /* Figure out how big the table can be with default button size */
  size = size2pixel(config.scene_size);
  across = (gdk_screen_width() - 50) / ( 2*size + 2*SPACING);
  down = (gdk_screen_height() - 50) / size - 1; /* -1 for the stop button */
  
  /* If this is not big enough, shrink the buttons */
  /* XXX FIXME XXX  if it still overflows, could put scroll bars around */
  if (across*down < count1) {
    double shrink = sqrt((double)(across*down) / (double)count1);
    size = floor(config.scene_size*shrink);
    if (size < 2) size = 2;
  }
  else size = config.scene_size;
  
  /* Construct a table of the same aspect ratio */
  rows = floor(sqrt((double)(count1*down) / (double)across));
  cols = count1/rows;
  rows = (count1 + cols-1)/cols;
  
  /* Display all the scene in a table */
  table = gtk_table_new(cols, rows, TRUE);
  j = 0;
  for (i = 0; i < count1; i++) {
      
    /* Find working scene corresponding to base scene */
    cmp = -1;
    while (j < count2) {
      cmp = dirsort(files1+i, files2+j);
      if (cmp > 0) free(files2[j++]);
      else break;
    }
    
    /* Add base scene or base/working scene pair to the table */
    base = select_scene_button (config.sticker_dir, files1[i], size);
    config_tip(base, base_tip);
    if (cmp == 0) {
      work = select_scene_button (config.work_dir, files2[j], size);
      config_tip(work, work_tip);
      entry = gtk_hbox_new(TRUE, 0);
      gtk_box_pack_start (GTK_BOX(entry), base, TRUE, TRUE, 0);
      gtk_box_pack_start (GTK_BOX(entry), work, TRUE, TRUE, 0);
      free(files2[j++]);
    }
    else {
      entry = base;
    }
    gtk_table_attach(GTK_TABLE(table), entry,
		     i/cols, i/cols+1, i%cols, i%cols+1,
		     GTK_EXPAND|GTK_FILL, GTK_EXPAND|GTK_FILL, SPACING, 0);
    
    free(files1[i]);
  }
  free(files1);
  while (j < count2) free(files2[j++]);
  if (count2) free(files2);

  return table;
}

void select_scene()
{
  GtkWidget *window, *vbox, *hbox, *table;

  TRACE("select scene");

  /* Set up select scene window */
  window = gtk_window_new(GTK_WINDOW_DIALOG);
  gtk_window_position (GTK_WINDOW (window), GTK_WIN_POS_CENTER);
  gtk_window_set_title (GTK_WINDOW (window), _("Sticker scene select"));

  gtk_signal_connect (GTK_OBJECT (window), "delete_event",
		      (GtkSignalFunc) select_scene_quit, NULL);
  gtk_signal_connect (GTK_OBJECT (window), "enter_notify_event",
		      (GtkSignalFunc) enter_window, NULL);
  gtk_signal_connect (GTK_OBJECT (window), "leave_notify_event",
		      (GtkSignalFunc) leave_window, NULL);
  gtk_signal_connect (GTK_OBJECT(window),"map_event",
		      (GtkSignalFunc) map_window, NULL);
  gtk_signal_connect (GTK_OBJECT (window), "key_press_event",
		      GTK_SIGNAL_FUNC (select_scene_keystroke), NULL);
  gtk_widget_set_events (window, GDK_EXPOSURE_MASK
			 | GDK_LEAVE_NOTIFY_MASK
			 | GDK_ENTER_NOTIFY_MASK
			 | GDK_KEY_PRESS_MASK);

  vbox = gtk_vbox_new (FALSE, 0);
  gtk_box_set_spacing (GTK_BOX (vbox), 10);
  gtk_container_border_width (GTK_CONTAINER (vbox), 0);
  gtk_container_add(GTK_CONTAINER(window), vbox);

  /* Add the scene table */
  table = build_scene_table();
  if (table == NULL) {
    table = gtk_label_new("\
No scenes available. Please restart stickers using\n\n\
 $ cd <stickers-src>; ./stickers scenes\n\n\
or type make install and run stickers directly.");
  }
  gtk_box_pack_start(GTK_BOX(vbox), table, TRUE, TRUE, 0);
    
  /* Add a stop/config buttons */
  gtk_box_pack_start (GTK_BOX (vbox), gtk_hseparator_new(), FALSE, FALSE, 0);
  hbox = gtk_hbox_new (FALSE, 0);
  gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
  add_control(hbox, "Stop", select_scene_quit, _("\
Exit stickers. Equivalent to Esc. Use \"!\" to enter configuration window."));
  if (config.config)
    add_control(hbox, "Configure", select_scene_config, _("\
Enter the configuration window. Equivalent to \"!\"."));

  gtk_widget_show_all(window);
  set_pointer (window);
  TRACE("About to select scene");
}
