/* Stickers.c
 *
 * Copyright 1999, Paul Kienzle (pkienzle@cs.indiana.edu)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * * Initial revision
 */

#define VERBOSE 0
#include <math.h>
#include <string.h>
#include "misc.h"
#include "canvas.h"
#include "stickers.h"

Sticker *sticker = NULL;

/* =============== Sticker rendering ==================================== */

/* Apply the colour map rotation to the sticker without retransforming
 * the scale or rotation.  Be sure that sticker_render has been called
 * first.  Note that sticker_render automatically calls sticker colour.
 * The only reason to call this is if you are changing the colour map
 * only, and not the size.
 *
 * Rotations are as follows:
 *   hue 0-360  goes once around the colour circle
 *   saturation, brightness, alpha
 *       -200  normal saturation
 *       -100  completely desaturated
 *          0  normal saturation
 *        100  completely saturated
 *        200  normal saturation
 *        300  completely desaturated
 *        400  normal saturation
 */
static double untangle(int v)
{
  /* convert up-down-up-down cycle into up range -100:100 */
  v = (v+400)%400;
  if (v > 300) v -= 400;
  if (v > 100) v = 200 - v;

  /* Return it as a percent */
  return (double)v/100.0;
}

/* Render the sticker as an image that can be pasted relatively easily
 * on to the image. Be sure to set all changes (scale, rotation, flip,
 * and colour) before calling. */
GdkImlibImage *sticker_render(Sticker *s)
{
  /* remember the most recent sticker */
  static gdouble scale=1.0;
  static gint rotate=0;
  static gboolean flip=FALSE;
  static GdkImlibImage *image=NULL;
  static GdkImlibImage *render=NULL;

  GdkImlibImage *colour;

  if (image != s->image || scale != s->scale*canvas.scale*s->base_scale || 
      rotate != s->rotate || flip != s->flip) {
    if (render) gdk_imlib_kill_image(render);
    render = gdk_imlib_transform_image(s->image,
				       s->scale*s->base_scale*canvas.scale, 
				       s->rotate*M_PI/180.0,
				       s->flip);
    scale = s->scale; rotate = s->rotate; flip = s->flip; image = s->image;
  }
  s->width = render->rgb_width; s->height = render->rgb_height;



  /* Finally, do the colour changes if any. */
  colour = gdk_imlib_clone_image(render);
  if (s->hue != 0 || s->saturation != 0 || 
      s->brightness != 0 || s->alpha != 0)
    gdk_imlib_hsv_rotate(colour, (double)(s->hue%360),
			 untangle(s->saturation),
			 untangle(s->brightness),
			 untangle(s->alpha));
  

  return colour;
}

/* ============== Sticker controls ================================ */
/* Generic sticker controls have a single parameter which specifies
 * the degree of movement (1 for fine, 2 for coarse), the direction
 * of movement (positive for forward, negative for reverse), or
 * the restoration of the default (0).  Each individual control
 * decides what is coarse movement and what is fine movement.
 */



void sticker_scale(gint c)
{
  gdouble scale, allscale;
  
  switch(c) {
  case  0: scale = 1.0/sticker->scale; break;
  case  1: scale = 1.01; break;
  case -1: scale = 1/1.01; break;
  case  2: scale = 1.10; break;
  case -2: scale = 1/1.10; break;
  default: scale = 1.0;
  }
  TRACE("Scaling sticker by %f.", scale);
  scale *= sticker->scale;

  /* Make sure the new sticker will be at least one pixel in each direction,
  * and won't be too big for the scene. */
  allscale = scale*sticker->base_scale*canvas.scale;
  if (allscale*sticker->image->rgb_width >= 1.0 &&
      allscale*sticker->image->rgb_height >= 1.0 &&
      allscale*sticker->image->rgb_width <= canvas.width &&
      allscale*sticker->image->rgb_height <= canvas.height) {
    sticker->scale = scale;
  }
}

void sticker_rotate(gint c)
{
  switch(c) {
  case  0: c = -sticker->rotate; break;
  case  1: c = 1; break;
  case -1: c = -1; break;
  case  2: c = 15; break;
  case -2: c = -15; break;
  }
  sticker->rotate = (sticker->rotate + c + 360) % 360;
}

void sticker_hue(gint c)
{
  switch(c) {
  case  0: c = -sticker->hue; break;
  case  1: c = 1; break;
  case -1: c = -1; break;
  case  2: c = 15; break;
  case -2: c = -15; break;
  }
  sticker->hue = (sticker->hue + c + 360) % 360;
}

void sticker_saturation(gint c)
{
  switch(c) {
  case  0: c = -sticker->saturation; break;
  case  1: c = 1; break;
  case -1: c = -1; break;
  case  2: c = 10; break;
  case -2: c = -10; break;
  }
  sticker->saturation = (sticker->saturation + c + 400) % 400;
}

void sticker_brightness(gint c)
{
  switch(c) {
  case  0: c = -sticker->brightness; break;
  case  1: c = 1; break;
  case -1: c = -1; break;
  case  2: c = 10; break;
  case -2: c = -10; break;
  }
  sticker->brightness = (sticker->brightness + c + 400) % 400;
}

void sticker_fade(gint c)
{
  switch(c) {
  case  0: c = -sticker->alpha; break;
  case  1: c = 1; break;
  case -1: c = -1; break;
  case  2: c = 10; break;
  case -2: c = -10; break;
  }
  sticker->alpha = (sticker->alpha + c + 400) % 400;
}

void sticker_vertical(gint control)
{
  gint x, y;
  g_return_if_fail(sticker!=NULL);
  
  if (!cursor.on_canvas) control = 0;
  x = cursor.x; y=cursor.y;
  switch (control) {
  case  0: x = canvas.width/2, y=canvas.height/2; break;
  case -1: y = MAX(cursor.y-1, 0); break;
  case  1: y = MIN(cursor.y+1, canvas.height); break;
  case -2: y = MAX(cursor.y-15, 0); break;
  case  2: y = MIN(cursor.y+15, canvas.height); break;
  }
  cursor.x = x; cursor.y = y; cursor.on_canvas = TRUE;
}

void sticker_horizontal(gint control)
{
  gint x, y;
  g_return_if_fail(sticker!=NULL);

  if (!cursor.on_canvas) control = 0;
  x = cursor.x; y = cursor.y;
  switch (control) {
  case  0: x = canvas.width/2, y=canvas.height/2; break;
  case -1: x = MAX(x-1, 0); break;
  case  1: x = MIN(x+1, canvas.width); break;
  case -2: x = MAX(x-15, 0); break;
  case  2: x = MIN(x+15, canvas.width); break;
  }
  cursor.x = x; cursor.y = y; cursor.on_canvas = TRUE;
}


void sticker_set_cursor()
{
  GdkImlibImage *image;

  /* XXX FIXME XXX : shouldn't this be in cursor_set? */
  if (cursor.image) gdk_imlib_kill_image(cursor.image);
  if (sticker) {
    image = sticker_render(sticker);
    cursor_set(image, image->rgb_width/2, image->rgb_height/2);
  }
}


void sticker_flip()
{
  sticker->flip = !sticker->flip;
  sticker_set_cursor();
}

void sticker_restore()
{
  sticker->hue = sticker->saturation = sticker->brightness = 0;
  sticker->alpha = 0;
  sticker->scale = 1.0;
  sticker->rotate = 0;
  sticker->flip = FALSE;
  sticker_set_cursor();
}


/* Define a new sticker from the given image --- called from main when a
 * new sticker is selected. */
void sticker_set(Sticker *s)
{
  TRACE("Changing sticker");
  sticker = s;
  sticker_set_cursor();
}

gboolean sticker_match(Sticker *a, Sticker *b)
{
  return a->image == b->image && a->scale == b->scale &&
    a->rotate == b->rotate && a->flip == b->flip &&
    a->x == b->x && a->y == b->y &&
    a->hue == b->hue && a->saturation == b->saturation &&
    a->alpha == b->alpha && a->brightness == b->brightness;
}

Sticker* sticker_new(GdkImlibImage *image)
{
  Sticker *s;

  s = g_new(Sticker, 1);
  memset(s, 0, sizeof(*s));
  s->base_scale = 1.0;
  s->scale = 1.0;
  s->image = image;
  return s;
}

void sticker_save(Sticker *s, FILE *file)
{
  char *filename;
  /* Obscure: must remove the sticker directory name from the path
   * before saving the sticker --- this means that we can't do relative
   * paths to stickers in our scene files. Oh, well! */
  filename = strrchr(s->image->filename, '/')+1;
  if (filename == NULL) filename = s->image->filename;
  fprintf(file, "n %s", filename);
  if (s->rotate != 0) 
    fprintf(file, " r %d", s->rotate);
  if (s->flip) 
    fprintf(file, " X");
  if (s->scale != 1.0)
    fprintf(file, " s %f", s->scale);
  if (s->hue!=0 || s->saturation!=0 || s->brightness!=0)
    fprintf(file, " c %d,%d,%d", 
	    s->hue, s->saturation, s->brightness);
  if (s->alpha!=0)
    fprintf(file, " a %d", s->alpha);
  fprintf(file, " p %.4f,%.4f", s->rel_x, s->rel_y);
  fprintf(file, "\n");
}

/* This reads the next sticker parameters from a file into the current
 * sticker, returning the name of the image --- the caller must set the
 * returned sticker's s->image variable to the actual image.
 */
Sticker* sticker_load(FILE *file, gchar **imagename)
{
  Sticker *s;
  gchar name[256], part;
  gint v;
  gboolean err = FALSE;

  s = sticker_new(NULL);
  name[0] = '\0';
  do {
    v = fscanf(file, " %c", &part);
    if (v != 1) break;
    switch(part) {
    case '@':
      ungetc(part, file);
      break;
    case 'n':
      v=fscanf(file, "%255s ", name);
      if (v != 1) err = TRUE;
      break;
    case 'r': 
      v=fscanf(file, "%d ", &s->rotate); 
      if (v != 1) err = TRUE;
      break;
    case 'X': s->flip=TRUE; break;
    case 's': 
      v=fscanf(file, "%lf ", &s->scale);
      if (v != 1) err = TRUE;
      break;
    case 'c': 
      v=fscanf(file, "%d, %d, %d ", 
	       &s->hue, &s->saturation, &s->brightness); 
      if (v != 3) err = TRUE;
      break;
    case 'a':
      v=fscanf(file, "%d ", &s->alpha);
      if (v != 1) err = TRUE;
      break;
    case 'p':
      v=fscanf(file, "%lf, %lf ", &s->rel_x, &s->rel_y);
      if (v != 2) err = TRUE;
      break;
    default:
      err = TRUE;
    }
  } while (part != '@');

  if (err) {
    g_free(s);
    return NULL;
  }
  else {
    *imagename = g_strdup(name);
    return s;
  }
}
